Package: dcortools
Title: Providing Fast and Flexible Functions for Distance Correlation
        Analysis
Description: Provides methods for distance covariance and distance correlation (Szekely, et al. (2007) <doi:10.1214/009053607000000505>), generalized version thereof (Sejdinovic, et al. (2013) <doi:10.1214/13-AOS1140>) and corresponding tests (Berschneider, Bottcher (2018) <doi:10.48550/arXiv.1808.07280>. Distance standard deviation methods (Edelmann, et al. (2020) <doi:10.1214/19-AOS1935>) and distance correlation methods for survival endpoints (Edelmann, et al. (2021) <doi:10.1111/biom.13470>) are also included.
Version: 0.2.1
Authors@R: c(person("Dominic", "Edelmann", email = "dominic.edelmann@dkfz-heidelberg.de", role = c("aut", "cre")), person("Jochen", "Fiedler", email = "jochen.fiedler@itwm.fraunhofer.de", role = c("aut")))
Depends: Rdpack, R (>= 3.3.3), Rcpp (>= 0.11.0), Rfast
Imports: ggplot2, pheatmap, Hmisc, stats, RSpectra, CompQuadForm,
        survey
RdMacros: Rdpack
LinkingTo: RcppArmadillo, Rcpp, RcppEigen
License: GPL-3
Encoding: UTF-8
NeedsCompilation: yes
Author: Dominic Edelmann [aut, cre],
  Jochen Fiedler [aut]
Maintainer: Dominic Edelmann <dominic.edelmann@dkfz-heidelberg.de>
Repository: CRAN
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-16 18:52:48 UTC; edelmand
Date/Publication: 2026-07-16 19:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,dcmatrix)
S3method(print,dcmatrix)
S3method(print,dctest)
export(dcmatrix)
export(dcorgaussianbiv)
export(dcsis)
export(distcor)
export(distcov)
export(distcov.test)
export(distsd)
export(hsplot)
export(ipcw.dcor)
export(ipcw.dcov)
export(ipcw.dcov.test)
importFrom(Rcpp,sourceCpp)
importFrom(Rdpack,reprompt)
importFrom(Rfast,standardise)
useDynLib(dcortools, .registration = TRUE)

dcortools R package 
####################
by Dominic Edelmann (dominic.edelmann@dkfz-heidelberg.de) and Jochen Fiedler

This package provides functionalities for calculating association measures and 
dependence tests based on distance covariance and distance correlation.

#' Calculates distance covariance and distance correlation matrices
#'
#' @param X A data.frame or matrix.
#' @param Y Either NULL or a data.frame or a matrix with the same number of rows as X. If only X is provided, distance covariances/correlations are calculated between all groups in X. If X and Y are provided, distance covariances/correlations are calculated between all groups in X and all groups of Y.
#' @param calc.dcov logical; specifies if the distance covariance matrix is calculated.
#' @param calc.dcor logical; specifies if the distance correlation matrix is calculated.
#' @param calc.cor If set as "pearson", "spearman" or "kendall", a corresponding correlation matrix is additionally calculated.
#' @param calc.pvalue.cor logical; IF TRUE, a p-value based on the Pearson or Spearman correlation matrix is calculated (not implemented for calc.cor ="kendall") using Hmisc::rcorr.
#' @param return.data logical; specifies if the dcmatrix object should contain the original data.
#' @param test specifies the type of test that is performed, "permutation" performs a Monte Carlo Permutation test. "gamma" performs a test based on a gamma approximation of the test statistic under the null. "conservative" performs a conservative two-moment approximation. "bb3" performs a quite precise three-moment approximation and is recommended when computation time is not an issue.
#' @param adjustp If setting this parameter to "holm", "hochberg", "hommel", "bonferroni", "BH", "BY" or "fdr", corresponding adjusted p-values are additionally returned for the distance covariance test.
#' @param b specifies the number of random permutations used for the permutation test. Ignored for all other tests.
#' @param affine logical; indicates if the affinely transformed distance covariance should be calculated or not.
#' @param standardize specifies if data should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.
#' @param bias.corr logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.
#' @param group.X A vector, each entry specifying the group membership of the respective column in X. Each group is handled as one sample for calculating the distance covariance/correlation matrices. If NULL, every sample is handled as an individual group.
#' @param group.Y A vector, each entry specifying the group membership of the respective column in Y. Each group is handled as one sample for calculating the distance covariance/correlation matrices. If NULL, every sample is handled as an individual group.
#' @param metr.X Either a single metric or a list providing a metric for each group in X (see examples).
#' @param metr.Y see metr.X.
#' @param use "all" uses all observations, "complete.obs" excludes NAs, "pairwise.complete.obs" uses pairwise complete observations for each comparison.
#' @param algorithm specifies the algorithm used for calculating the distance covariance. 
#' 
#' "fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 
#' 
#' "memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 
#' 
#' "standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.
#' 
#' "auto" chooses the best algorithm for the specific setting using a rule of thumb.
#' 
#' "memsave" is typically very inefficient for dcmatrix and should only be applied in exceptional cases.
#' 
#' @param fc.discrete logical; If TRUE, "discrete" metric is applied automatically on samples of type "factor" or "character".
#' @param calc.dcov.pw logical; If TRUE, a distance covariance matrix between the univariate observations/columns is additionally calculated. Not meaningful if group.X and group.Y are not specified.
#' @param calc.dcor.pw logical; If TRUE, a distance correlation matrix between the univariate observations/columns is additionally calculated. Not meaningful if group.X and group.Y are not specified.
#' @param test.pw specifies a test (see argument "test") that is performed between all single observations.
#' @param metr.pw.X Either a single metric or a list providing a metric for each single observation/column in X (see metr.X).
#' @param metr.pw.Y See metr.pw.Y.
#' @return S3 object of class "dcmatrix" with the following components
#' \item{name X, Y}{description original data (if return.data = TRUE).} 
#' \item{name dcov, dcor}{distance covariance/correlation matrices between the groups specified in group.X/group.Y (if calc.dcov/calc.dcor = TRUE).} 
#' \item{name corr}{correlation matrix between the univariate observations/columns (if cal.cor is "pearson", "spearman" or "kendall").}
#' \item{name pvalue}{matrix of p-values based on a corresponding distance covariance test based on the entries in dcov (if argument test is not "none").} 
#' \item{name pvalue.adj}{matrix of p-values adjusted for multiple comparisons using the method specified in argument adjustp.} 
#' \item{name pvalue.cor}{matrix of pvalues based on "pearson"/"spearman" correlation (if calc.cor is "pearson" or "spearman" and calc.pvalue.cor = TRUE).}
#' \item{name dcov.pw,dcor.pw}{distance covariance/correlation matrices between the univariate observations (if calc.dcov.pw/calc.dcor.pw = TRUE.)} 
#' \item{name pvalue.pw}{matrix of p-values based on a corresponding distance covariance test based on the entries in dcov.pw (if argument test is not "none").} 
#' @export
#' @references
#' \insertRef{berschneider2018complex}{dcortools}
#' 
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huang2017statistically}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#'
#'@examples
#' X <- matrix(rnorm(1000), ncol = 10)
#' 
#' dcm <- dcmatrix(X, test="bb3",calc.cor = "pearson",
#'  calc.pvalue.cor = TRUE, adjustp = "BH") 
#'  
#' dcm <- dcmatrix(X, test="bb3",calc.cor = "pearson", 
#'  calc.pvalue.cor = TRUE, adjustp = "BH", 
#'  group.X = c(rep(1, 5), rep(2, 5)), 
#'  calc.dcor.pw = TRUE, test.pw = "bb3")
#' 
#' 
#' Y <- matrix(rnorm(600), ncol = 6)
#' 
#' Y[,6] <- rbinom(100, 4, 0.3)
#' 
#' dcm <- dcmatrix(X, Y, test="bb3",calc.cor = "pearson",
#'  calc.pvalue.cor = TRUE, adjustp = "BH")
#'   
#' dcm <- dcmatrix(X, Y, test="bb3",calc.cor = "pearson",
#'  calc.pvalue.cor = TRUE, adjustp = "BH",
#'  group.X = c(rep("group1", 5), rep("group2", 5)),
#'  group.Y = c(rep("group1", 5), "group2"), 
#'  metr.X = "gaussauto",
#'  metr.Y = list("group1" = "gaussauto", "group2" = "discrete"))
dcmatrix <- function (X,
                      Y = NULL,
                      calc.dcov = TRUE,
                      calc.dcor = TRUE,
                      calc.cor = "none",
                      calc.pvalue.cor = FALSE,
                      return.data = TRUE,
                      test = "none",
                      adjustp = "none",
                      b = 499,
                      affine = FALSE,
                      standardize = FALSE,
                      bias.corr = FALSE,
                      group.X = NULL,
                      group.Y = NULL,
                      metr.X = "euclidean",
                      metr.Y = "euclidean",
                      use="all",
                      algorithm ="auto",
                      fc.discrete = FALSE,
                      calc.dcor.pw = FALSE,
                      calc.dcov.pw = FALSE,
                      test.pw = "none",
                      metr.pw.X = "euclidean",
                      metr.pw.Y = "euclidean"
                      ) {
  
  output <- .dcmatrixmain(X, Y, calc.dcov, calc.dcor, calc.cor, calc.pvalue.cor, return.data, test, adjustp, b, affine, standardize, bias.corr, group.X, group.Y, metr.X, metr.Y, use, algorithm, fc.discrete)
  output$call <- match.call()
  
  if (calc.dcor.pw | calc.dcov.pw | test.pw != "none") {
    output2 <- .dcmatrixmain(X, Y, calc.dcov = calc.dcov.pw, calc.cor = "no", calc.dcor = calc.dcor.pw, calc.pvalue.cor = FALSE, return.data, test = test.pw, adjustp, b, affine, standardize, bias.corr, group.X = NULL, group.Y = NULL, metr.X = metr.pw.X, metr.Y = metr.pw.Y, use, algorithm, fc.discrete)
    output$dcor.pw <- output2$dcor
    output$dcov.pw <- output2$dcov
    output$pvalue.pw <- output2$pvalue
    output$metr.pw.X <- metr.pw.X
    output$metr.pw.Y <- metr.pw.Y
    output$dX.pw <-  output2$dX
    output$dY.pw <-  output2$dY
    names.X <- names.Y <- colnames(X)
    if (!is.null(Y))
      names.Y <- colnames(Y)
    if (!is.null(output$dcor.pw)) {
      rownames(output$dcor.pw) <- names.X
      colnames(output$dcor.pw) <- names.Y
    }
    if (!is.null(output$dcov.pw)) {
      rownames(output$dcov.pw) <- names.X
      colnames(output$dcov.pw) <- names.Y
    }
    
    if (!is.null(output$pvalue.pw)) {
      rownames(output$pvalue.pw) <- names.X
      colnames(output$pvalue.pw) <- names.Y
    }
    
    
  }
  
  output$test.pw <- test.pw
  output$calc.dcor.pw <- calc.dcor.pw
  output$calc.dcov.pw <- calc.dcov.pw
  
  return(output)
  
}


.dcmatrixmain <- function (X,
                           Y = NULL,
                           calc.dcov = TRUE,
                           calc.dcor = TRUE,
                           calc.cor = "none",
                           calc.pvalue.cor = FALSE,
                           return.data = TRUE,
                           test = "none",
                           adjustp = "none",
                           b = 499,
                           affine = FALSE,
                           standardize=FALSE,
                           bias.corr = FALSE,
                           group.X = NULL,
                           group.Y = NULL,
                           metr.X = "euclidean",
                           metr.Y = "euclidean",
                           use="all",
                           algorithm ="auto",
                           fc.discrete = FALSE) {
  output <- list()
  
  
  if(return.data) {
    output$X <- X
    output$Y <- Y
  } else {
    output <- NULL
  }
  
  
  withY <- ifelse(is.null(Y), FALSE, TRUE)
  
  
  dogamma <- docons <- dobb3 <- doperm <- donotest <- FALSE
  
  if (test == "none")
    donotest <- TRUE 
  else if (test == "gamma")
    dogamma <- TRUE 
  else if (test == "conservative")
    docons <- TRUE
  else if (test == "bb3")
    dobb3 <- TRUE
  else if (test == "permutation")
    doperm <- TRUE
  else
    stop ("Test must be one of \"none\", \"permutation\", \"gamma\", \"bb3\" or \"conservative\"")
  
  
  use.all <- use.pw <- FALSE
  
  if (use == "complete.obs") {
    cc <- which(stats::complete.cases(X))
    if (withY) {
      cc <- intersect(cc,which(stats::complete.cases(Y)))
      Y <- Y[cc,]
    }
    X <- X[cc,]
    use.all <- TRUE
  } else if (use == "all") {
    use.all <- TRUE
  } else if (use == "pairwise.complete.obs") {
    use.pw <- TRUE
  } else {
    stop("use must be one of \"all\", \"complete.obs\" or \"pairwise.complete.obs\"")
  }
  
  
  
  
  if (is.vector(X)) {
    X <- as.matrix(X)
  }
  
  p <- ncol(X)
  n <- nrow(X)
  
  if (is.null(group.X)) {
    if (is.null(colnames(X)))
     group.X <- 1:p
     else
     group.X <- colnames(X)
  } 
  
  names.X <- names.Y <-  unique(group.X)
  dX <- dY <- length(unique(group.X))
  ms.grpX <- ms.grpY <- NULL
  groupslistX <- lapply(1:dX, function(t) which(group.X == names.X[t]))
  pX <- sapply(1:dX, function(t) length(groupslistX[[t]]))
  prepX <- as.list(rep(NA,dX))
  dvarX <- rep(NA,dX)
    
    
  # 
  # tblX <- table(group.X)
  # labelsX <- names.X <- names.Y <- as.factor(names(tblX))
  # #if (sum(group.X - 1:p) == 0)
  #  # names.X <- names.Y <- colnames(X)
  # pX <- as.numeric(tblX)
  # dX <- dY <- length(pX)
  # ms.grpX <- ms.grpY <- NULL
  # groupslistX <- lapply(1:dX, function(t) which(group.X == labelsX[t]))
  #prepX <- as.list(rep(NA,dX))
  #dvarX <- rep(NA,dX)
  

  
  lmX <- length(metr.X)
  if (lmX ==1) {
    metr.X <- as.list(replicate(dX, metr.X))
  } else if (lmX == 2) {
    ischar <- suppressWarnings(is.na(as.numeric(metr.X[2])))
    if (!ischar)
      metr.X <- lapply(1:dX, function(u) metr.X)
  }
  
  if (is.character(metr.X) & lmX == dX) {
    ischar <- suppressWarnings(is.na(as.numeric(metr.X[2])))
    if (ischar)
      metr.X <- as.list(metr.X)
  }
  
  if (use.all) {
    ms.X <- sapply(1:dX, function(t) any(!stats::complete.cases(X[,t])))
    ms.grpX <- which(ms.X)
  } else {
    ms.X <- rep(FALSE,dX)
    ms.grpX <- numeric(0)
  }
  
  ## normalize samples if calculation of affinely invariant distance covariance is desired
  if (affine) {
    for (j in 1 : dX) {
      if (use.all) {
        X[,groupslistX[[j]]] <- normalize.sample(X[,groupslistX[[j]]], n, pX[j])
      } else {
        cc <- stats::complete.cases(X[,groupslistX[[j]]])
        ncc <- length(cc)
        X[cc,groupslistX[[j]]] <- normalize.sample(X[cc,groupslistX[[j]]], n, pX[j])
      }
    }
  } else if (standardize) {
    if (use.all) {
      X[,groupslistX[[j]]] <- scale.sample(X[,groupslistX[[j]]], n, pX[j])
    } else {
      cc <- stats::complete.cases(X[,groupslistX[[j]]])
      ncc <- length(cc)
      X[cc,groupslistX[[j]]] <- scale.sample(X[cc,groupslistX[[j]]], n, pX[j])
    }
  }
  
  
  
  
  if (withY) {
    
    lmY <- length(metr.Y)
    
    
    if (is.vector(Y)) {
      Y <- as.matrix(Y)
    }
    
    q <- ncol(Y)
    m <- nrow(Y)
    
    if (is.null(group.Y)) {
      if (is.null(colnames(Y)))
        group.Y <- 1 : q
        else
        group.Y <- colnames(Y)
    }
   #   names.Y <- colnames(Y)
   # } else 
   #   names.Y <- unique(group.Y)
    
    names.Y <- unique(group.Y)
    dY <- length(unique(group.Y))
    groupslistY <- lapply(1:dY, function(t) which(group.Y == names.Y[t]))
    pY <- sapply(1:dY, function(t) length(groupslistY[[t]]))
    prepY <- as.list(rep(NA,dY))
    dvarY <- rep(NA,dY)
    
    #tblY <- table(group.Y)
    #labelsY <- as.factor(names(tblY))
    #pY <- as.numeric(tblY)
    #dY <- length(pY)
    #groupslistY <- lapply(1:dY, function(t) which(group.Y == labelsY[t]))
    #prepY <- as.list(rep(NA,dY))
    #dvarY <- rep(NA,dY)
    if (use.all) {
      ms.Y <- sapply(1:dY, function(t) any(!stats::complete.cases(Y[,t])))
      ms.grpY <- which(sapply(1:dY, function(t) any(!stats::complete.cases(Y[,t]))))
    } else {
      ms.Y <- rep(FALSE,dY)
      ms.grpY <- numeric(0)
    }
    
    if (lmY ==1) {
      metr.Y <- as.list(replicate(dY, metr.Y))
    } else if (lmX == 2) {
      ischar <- suppressWarnings(is.na(as.numeric(metr.Y[2])))
      if (!ischar)
        metr.Y <- lapply(1:dY, function(u) metr.Y)
    }
    
    if (is.character(metr.Y) & lmY == dY) {
      ischar <- suppressWarnings(is.na(as.numeric(metr.Y[2])))
      if (ischar)
        metr.Y <- as.list(metr.Y)
    }
    
    if (m != n) 
      stop("X and Y must have same number of rows (samples)")
    
    if (affine) {
      for (j in 1 : dY) {
        if (use.all) {
          Y[,groupslistY[[j]]] <- normalize.sample(Y[,groupslistY[[j]]], n, pY[j])
        } else {
          cc <- stats::complete.cases(Y[,groupslistY[[j]]])
          ncc <- length(cc)
          Y[cc,groupslistX[[j]]] <- normalize.sample(Y[cc,groupslistX[[j]]], ncc, pY[j])
        }
      }
    } else if (standardize) {
      if (use.all) {
        Y[,groupslistY[[j]]] <- scale.sample(Y[,groupslistY[[j]]], n, pY[j])
      } else {
        cc <- stats::complete.cases(X[,groupslistX[[j]]])
        ncc <- length(cc)
        Y[cc,groupslistY[[j]]] <- scale.sample(Y[cc,groupslistY[[j]]], n, pY[j])
      }
    }
    
  }
  
  
  
  
  if (algorithm == "auto") {
    gofast <- (((p == length(names.X))) * (n>200)) & (!dobb3) * all(metr.X %in% c("euclidean", "discrete"))
    if (withY) 
      gofast <- gofast * (q == length(names.Y)) * all(metr.Y %in% c("euclidean", "discrete"))
    if (gofast) {
      algorithm <- "fast"
    } else {
      algorithm <- "standard"
    }
  }
  
  alg.fast <- alg.standard <- alg.memsave <- FALSE
  
  if (algorithm == "fast") {
    alg.fast <- TRUE 
    if (doperm) 
      terms.smp <- function(terms, smp, ndisc = NULL) {sampleterms.fast.matr(terms, smp)}
  } else if (algorithm == "standard") {
    alg.standard <- TRUE
    if (doperm) 
      terms.smp <- function(terms, smp, ndisc = NULL) {sampleterms.standard(terms, smp)}
  }  else if (algorithm == "memsave") {
    alg.memsave <- TRUE
    if (doperm) 
      terms.smp <- function(terms, smp, ndisc = NULL) {sampleterms.memsave(terms, smp)}
  } else
    stop ("Algorithm must be one of \"fast\", \"standard\", \"memsave\" or \"auto\"")
  
  
  if (!alg.standard & dobb3) 
    stop("bb3 p-value calculation is only possible with algorithm=\"standard\"!")
  
  
  if (bias.corr == TRUE) {
    termstodcov2 <- function(aijbij,Sab,Tab,n) {
      aijbij/ n / (n - 3) - 2 * Sab / n / (n - 2) / (n - 3) + Tab / n / (n - 1) / (n - 2) / (n - 3) 
    }
    dcov2todcov <- function(dcov2) {
      sqrt(abs(dcov2)) * sign(dcov2)
    }
    dcov2todcor <- function(dcov2, dvarX, dvarY) {
      (sqrt(abs(dcov2)) * sign(dcov2)) / sqrt(sqrt(dvarX * dvarY))
    }
  } else  {
    termstodcov2 <- function(aijbij, Sab, Tab, n) {
      aijbij / n / n - 2 * Sab / n / n / n + Tab / n / n / n / n
    }
    dcov2todcov <- function(dcov2) {
      sqrt(dcov2)
    }
    dcov2todcor <- function(dcov2, dvarX, dvarY) {
      sqrt(dcov2) / sqrt(sqrt(dvarX * dvarY))
    }
  }
  
  if (dogamma) {
    testfunc <- function(terms, moms.X, moms.Y, dcov2, smp, ...) {
      n <- terms$ncc
      Saa <- vector_prod_sum(terms$aidot,terms$aidot)
      Sbb <- vector_prod_sum(terms$bidot,terms$bidot)
      Sab <- vector_prod_sum(terms$aidot,terms$bidot)
      dvarX <- terms$aijaij / n / (n - 3) - 2 * Saa/ n / (n - 2) / (n - 3) + terms$adotdot * terms$adotdot / n / (n - 1) / (n - 2) / (n - 3) 
      dvarY <- terms$bijbij / n / (n - 3) - 2 * Sbb / n / (n - 2) / (n - 3) + terms$bdotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3) 
      dcov2 <- terms$aijbij / n / (n - 3) - 2 * Sab / n / (n - 2) / (n - 3) + terms$adotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3) 
      U1 <- dvarX  * dvarY
      U2 <- terms$adotdot / n / (n - 1)
      U3 <- terms$bdotdot / n / (n - 1)
      alph <- 1 / 2 * (U2 ^ 2 * U3 ^ 2) / U1
      beta <- 1 / 2 * (U2 * U3) / U1
      stat <- n *  dcov2 + U2 * U3
      pval <- stats::pgamma(stat, alph, beta, lower.tail = FALSE) 
      return(pval)
    }
  } else if (doperm) {
    testfunc <- function(terms, moms.X, moms.Y, dcov2, smp,  ...) {
      n <- terms$ncc
      if (is.na(dcov2))
        return(NA)
      Tab <- terms$adotdot * terms$bdotdot
      reps <- lapply(1:b, function(t) {
        terms.sample <- terms.smp(terms,smp[[t]])
        return(termstodcov2(terms.sample$aijbij, terms.sample$Sab, Tab, n))
      })
      pval <- (1 + length(which(reps >= dcov2))) / (1 + b)
      return(pval)
    }
  } else if (docons) {
    testfunc <- function(terms, moms.X, moms.Y, dcov2, smp,...) {
      n <- terms$ncc
      est.m2 <- sum((moms.X * moms.Y)) / n ^ 10
      est.m1 <- terms$adotdot * terms$bdotdot / n ^ 3 / (n - 1)
      est.var <- (est.m2 - est.m1 ^ 2)
      alpha <- sqrt(est.var / 2 / est.m1 ^ 2)
      stat <- terms$aijbij / n - 2 * vector_prod_sum(terms$aidot,terms$bidot) / n ^ 2 + terms$adotdot * terms$bdotdot / n ^ 3
      pval <- stats::pchisq(stat * sqrt(2) / sqrt(est.var), df = 1 / alpha, lower.tail = FALSE)  
      return(pval)
    }
  } else if (dobb3) {
    testfunc <- function(terms, moms.X, moms.Y, dcov2, smp, ...) {
      n <- terms$ncc
      est.m2 <- sum((moms.X$vc * moms.Y$vc)) / n ^ 10
      est.m1 <- terms$adotdot * terms$bdotdot / n ^ 3 / (n - 1)
      est.var <- (est.m2 - est.m1 ^ 2)
      est.skw <- moms.X$skw * moms.Y$skw
      beta <- est.skw / sqrt(8)
      stat <- terms$aijbij / n - 2 * vector_prod_sum(terms$aidot,terms$bidot) / n ^ 2 + terms$adotdot * terms$bdotdot / n ^ 3
      centstat <- (stat - est.m1) /  sqrt(est.var)
      pval <- stats::pchisq((centstat * sqrt(2) + 1 / beta) / beta , df = 1 / beta ^ 2, lower.tail = FALSE)  
      return(pval)
    } 
  } else if (donotest) {
    testfunc <- function(terms, moms.X, moms.Y, dcov2, smp, ...) {}
  }
  
  if (!calc.dcov) {
    dcov2todcov <- function(dcov2) {}
  }
  
  if (!calc.dcor) {
    dcov2todcor <- function(dcov2, dvarX, dvarY) {}
  }
  
  if (doperm & use.all) {
    perms <- lapply(1:b, function(t) sample(1:n))
  } else {
    perms <- NULL
  }
  
  extendoutput <- doperm| ((dobb3|docons)*use.pw)
  
  
  
  if (fc.discrete) {
    for (j in 1:dX) {
      if (is.factor(X[,groupslistX[[j]]]) | is.character(X[,groupslistX[[j]]]))
        metr.X[[j]] <- "discrete"
    }
    if (withY) {
      for (j in 1:dY) {
        if (is.factor(Y[,groupslistY[[j]]]) | is.character(Y[,groupslistY[[j]]]))
          metr.Y[[j]] <- "discrete"
      }
    }
  }
  
  
  
  if (calc.cor %in% c("spearman","kendall", "pearson")) {
    output$corr <- stats::cor(X,Y, use = use, method = calc.cor)
    if (calc.pvalue.cor) {
      if (calc.cor %in% c("spearman", "pearson")) {
        if (!withY) {
          corrp <- Hmisc::rcorr(X, type = calc.cor)
          output$pvalue.cor <- corrp$P
          diag(output$pvalue.cor) <- 0
          if (use.all)
            output$pvalue.cor[which(corrp$n<n,arr.ind=TRUE)] <- NA
        } else {
          corrp <- Hmisc::rcorr(X,Y)
          output$pvalue.cor <- corrp$P[1:dX,(dX+1):(dX+dY)]
          if (use.all)
            output$pvalue.cor[which(corrp$n[1:dX,(dX+1):(dX+dY)]<n,arr.ind=TRUE)] <- NA
        }  
      } else 
        warning("P-Value calculation for Kendall correlation not implemented")
    }
  }
  
  
  
  if (alg.fast) {
    discrete.X <- (metr.X == "discrete")
    if (withY)
      discrete.Y <- (metr.Y == "discrete")
  }
  
  
  
  if (calc.dcov) {
    output$dcov <- matrix(nrow = dX, ncol = dY)
    rownames(output$dcov) <- names.X
    colnames(output$dcov) <- names.Y
  }
  
  if (calc.dcor) {
    output$dcor <- matrix(nrow = dX, ncol = dY)
    rownames(output$dcor) <- names.X
    colnames(output$dcor) <- names.Y
    if (!withY)
      diag(output$dcor) <- 1
  }
  
  if (!donotest) {
    output$pvalue <- matrix(nrow = dX, ncol = dY)
    rownames(output$pvalue) <- names.X
    colnames(output$pvalue) <- names.Y
    if (!withY)
      diag(output$pvalue) <- 0
  }
  
  momsX <- momsY <- NULL
  
  if ((docons | dobb3) & !use.pw) {
    momsX <- as.list(rep(NA,dX))
    if (withY)
      momsY <- as.list(rep(NA,dY))
  }
  
  
  
  for (j in setdiff(1:dX,ms.grpX)) {
    if (alg.fast) {
      prepX[[j]] <- prep.fast(X[,j], n, discrete = discrete.X[j], pairwise = use.pw)
    } else if (alg.memsave) {
      prepX[[j]] <- prep.memsave(X[,groupslistX[[j]]], n, pX[j],  metr.X = metr.X[[j]], pairwise = use.pw)
    } else if (alg.standard) {
      prepX[[j]] <- prep.standard(X[,groupslistX[[j]]], n, pX[j],  metr.X = metr.X[[j]], pairwise = use.pw)
    }  
    
    Saa <- vector_prod_sum(prepX[[j]]$aidot,prepX[[j]]$aidot)
    
    if ((docons | dobb3) & !use.pw) {
      momsX[[j]] <- calcmom(aijaij = prepX[[j]]$aijaij, Saa = Saa, adotdot = prepX[[j]]$adotdot, aidot = prepX[[j]]$aidot, distX = prepX[[j]]$distX, n = n, dobb3 = dobb3)
    }
    dvarX[j] <- termstodcov2(prepX[[j]]$aijaij, Saa, prepX[[j]]$adotdot*prepX[[j]]$adotdot, prepX[[j]]$ncc) 
  }
  
  if (!withY & calc.dcov) 
    diag(output$dcov) <- sqrt(dvarX)
  
  
  if (withY) {
    for (j in setdiff(1:dY,ms.grpY)) {
      if (alg.fast) {
        prepY[[j]] <- prep.fast(Y[,j], n, discrete = discrete.Y[j], pairwise = use.pw)
      } else if (alg.memsave) {
        prepY[[j]] <- prep.memsave(Y[,groupslistY[[j]]], n, pY[j], metr.X = metr.Y[[j]], pairwise = use.pw)
      } else if (alg.standard) {
        prepY[[j]] <- prep.standard(Y[,groupslistY[[j]]], n, pY[j], metr.X = metr.Y[[j]], pairwise = use.pw)
      }  
      Sbb <- vector_prod_sum(prepY[[j]]$aidot, prepY[[j]]$aidot)
      
      if ((docons | dobb3) & !use.pw) {
        momsY[[j]] <- calcmom(aijaij = prepY[[j]]$aijaij, Saa = Sbb, adotdot = prepY[[j]]$adotdot, aidot = prepY[[j]]$aidot, distX = prepY[[j]]$distX, n = n, dobb3 = dobb3)
      }
      dvarY[j] <- termstodcov2(prepY[[j]]$aijaij, Sbb, prepY[[j]]$adotdot*prepY[[j]]$adotdot, prepY[[j]]$ncc) 
    }
  }
  
  
  if (!withY) {
    if (dX > 1) {
      for (i in setdiff(1:(dX-1),ms.grpX)) {
        for (j in setdiff((i+1):dX,ms.grpX)) {
          if (alg.fast) {
            terms <- preptoterms.fast(prepX[[i]], prepX[[j]], n, pairwise = use.pw, discrete.X[[i]], discrete.X[[j]], perm = extendoutput)
          } else if (alg.memsave) {
            terms <- preptoterms.memsave(prepX[[i]], prepX[[j]], metr.X[[i]], metr.X[[j]], n, pairwise = use.pw, perm = extendoutput) 
          } else if (alg.standard) {
            terms <- preptoterms.standard(prepX[[i]], prepX[[j]], n, pairwise = use.pw, perm = extendoutput)
          }  
          dcov2XY <- termstodcov2(terms$aijbij, vector_prod_sum(terms$aidot, terms$bidot), terms$adotdot * terms$bdotdot, terms$ncc)
          output$dcov[i,j] <- output$dcov[j,i] <- dcov2todcov(dcov2 = dcov2XY)
          if (use.pw) {
            Saa <- vector_prod_sum(terms$aidot, terms$aidot)
            Sbb <- vector_prod_sum(terms$bidot, terms$bidot)
            dvX <-  termstodcov2(terms$aijaij, Saa, terms$adotdot * terms$adotdot, terms$ncc)
            dvY <-  termstodcov2(terms$bijbij, Sbb, terms$bdotdot * terms$bdotdot, terms$ncc)
            if (docons | dobb3) {
              moms.X <- calcmom(aijaij = terms$aijaij, Saa = Saa, adotdot = terms$adotdot, distX = terms$distX, n = terms$ncc, aidot = terms$aidot, dobb3 = dobb3)
              moms.Y <- calcmom(aijaij = terms$bijbij, Saa = Sbb, adotdot = terms$bdotdot, distX = terms$distY, n = terms$ncc, aidot = terms$bidot, dobb3 = dobb3)
            }
            if (doperm) {
              perms <- lapply(1:b, function(t) sample(1:terms$ncc))
            }
            
          } else {
            dvX <- dvarX[i]
            dvY <- dvarX[j]
            if (docons | dobb3) {
              moms.X <- momsX[[i]]
              moms.Y <- momsX[[j]]
            }
          }
          
          
          
          output$dcor[i,j] <- output$dcor[j,i] <- dcov2todcor(dcov2 = dcov2XY, dvX, dvY)
          
          
          
          output$pvalue[i,j] <- output$pvalue[j,i] <- testfunc(dcov2 = dcov2XY, terms = terms, moms.X = moms.X, moms.Y = moms.Y, n = n, smp = perms, prepX[[i]], prepX[[j]])
        }
      }
    }
  } else {
    for (i in setdiff(1:dX,ms.grpX)) {
      for (j in setdiff(1:dY,ms.grpY)) {
        if (alg.fast) {
          terms <- preptoterms.fast(prepX[[i]], prepY[[j]], n, pairwise = use.pw, discrete.X[[i]], discrete.Y[[j]], perm = extendoutput)
        } else if (alg.memsave) {
          terms <- preptoterms.memsave(prepX[[i]], prepY[[j]], metr.X[[i]], metr.Y[[j]], n, pairwise = use.pw, perm = extendoutput) 
        } else if (alg.standard) {
          terms <- preptoterms.standard(prepX[[i]], prepY[[j]], n, pairwise = use.pw, perm = extendoutput)
        }  
        dcov2XY <- termstodcov2(terms$aijbij, vector_prod_sum(terms$aidot, terms$bidot), terms$adotdot * terms$bdotdot, terms$ncc)
        output$dcov[i,j] <- dcov2todcov(dcov2 = dcov2XY)
        if (use.pw) {
          Saa <- vector_prod_sum(terms$aidot, terms$aidot)
          Sbb <- vector_prod_sum(terms$bidot, terms$bidot)
          dvX <-  termstodcov2(terms$aijaij, Saa, terms$adotdot * terms$adotdot, terms$ncc)
          dvY <-  termstodcov2(terms$bijbij, Sbb, terms$bdotdot * terms$bdotdot, terms$ncc)
          if (docons | dobb3) {
            moms.X <- calcmom(aijaij = terms$aijaij, Saa = Saa, adotdot = terms$adotdot, distX = terms$distX, aidot = terms$aidot, n = terms$ncc, dobb3 = dobb3)
            moms.Y <- calcmom(aijaij = terms$bijbij, Saa = Sbb, adotdot = terms$bdotdot, distX = terms$distY, aidot = terms$bidot, n = terms$ncc, dobb3 = dobb3)
          }
          if (doperm) {
            perms <- lapply(1:b, function(t) sample(1:terms$ncc))
          }
        } else {
          dvX <- dvarX[i]
          dvY <- dvarY[j]
          if (docons | dobb3) {
            moms.X <- momsX[[i]]
            moms.Y <- momsY[[j]]
          }
        }
        output$dcor[i,j] <- dcov2todcor(dcov2 = dcov2XY, dvX, dvY)
        
        
        output$pvalue[i,j] <- testfunc(dcov2 = dcov2XY, terms = terms, moms.X = moms.X, moms.Y = moms.Y, smp = perms, prepX[[i]], prepY[[j]])
      }
    }
  }
  
  
  
  
  
  if (adjustp %in% c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr")) {
    if (withY) {
      output$pvalue.adj <- matrix(stats::p.adjust(output$pvalue,method = adjustp), ncol = dY)
    } else {
      ind <- which(lower.tri(output$pvalue), arr.ind=TRUE)
      pvec <- as.vector(output$pvalue[ind])
      pvec <- stats::p.adjust(pvec, method = adjustp)
      output$pvalue.adj <- diag(0,dX)
      ind2 <- ind[,2:1]
      output$pvalue.adj[ind] <- output$pvalue.adj[ind2] <- pvec
    }  
  } else if (adjustp != "none")
    warning ("adjustp should be one of \"holm\", \"hochberg\", \"hommel\", \"bonferroni\", \"BH\", \"BY\", \"fdr\" \n
               No p-value correction performed")
  
  class(output) <- "dcmatrix"
  output$withY <- withY
  output$dX <- dX
  output$n <- n
  output$b <- b
  output$test <- test
  output$calc.dcov <- calc.dcov
  output$calc.dcor <- calc.dcor
  output$bias.corr <- bias.corr
  output$affine <- affine
  output$calc.cor <- calc.cor
  output$group.X  <-  group.X
  output$names.X <-  names.X
  output$groupslistX <-  groupslistX
  
  if (withY) {
    output$group.Y <- group.Y
    output$dY <- dY
    output$names.Y <- names.Y
    output$groupslistY <- groupslistY
  } 
  
  return(output)

}

#' Performs a distance covariance test.
#'
#' @param X contains either the first sample or its corresponding distance matrix.
#'
#' In the first case, X can be provided either as a vector (if one-dimensional), a matrix or a data.frame (if two-dimensional or higher). 
#'
#' In the second case, the input must be a distance matrix corresponding to the sample of interest.
#'
#' If X is a sample, type.X must be specified as "sample". If X is a distance matrix, type.X must be specified as "distance".
#' @param Y see X.
#' @param method specifies the type of test that is performed.
#'  
#' "permutation" performs a Monte Carlo Permutation test. 
#' 
#' "gamma" performs a test based on a gamma approximation of the test statistic under the null \insertCite{huang2017statistically}{dcortools}. This test tends to be anti-conservative, if the ``real'' p-value is small
#' 
#' "conservativebb" performs an (asymptotically) conservative two-moment approximation \insertCite{berschneider2018complex}{dcortools}.
#' 
#' "conservativesz" performs the (asymptotically) conservative procedure proposed by \insertCite{szekely2007}{dcortools}.
#' 
#' "spv" performs the Chi-Square Test of Distance correlation proposed by Shen, Panda and Vogelstein (2021).
#' 
#' "spectral" performs the spectral approach based on the asymptotic distribution.
#'   
#' "bb3" performs a  three-moment approximation  \insertCite{berschneider2018complex}{dcortools}. This is the most precise parametric option, but only available with the standard algorithm.
#' 
#' "wildbs1" and "wildbs2" perform wild bootstrap tests \insertCite{chwialkowski2014wild}{dcortools}; experimental at the moment.
#' @param b integer; specifies the number of random permutations/bootstrap samples used for the permutation or wild bootstraps tests. Ignored for other tests.
#' @param ln numeric; block size parameter for wild bootstrap tests. Ignored for other tests.
#' @param affine logical; specifies if the affinely invariant distance covariance \insertCite{dueck2014affinely}{dcortools} should be calculated or not.
#' @param standardize logical; specifies if X and Y should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.
#' @param bias.corr logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.
#' @param type.Y see type.X.
#' @param metr.X specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").
#' 
#'  Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
#'  
#'  For "alpha", "minkowski", "gauss", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric, parameter)", c("gaussian", 3) for example uses a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
#'  
#'  See \insertCite{lyons2013distance,sejdinovic2013equivalence,bottcher2017detecting;textual}{dcortools} for details.
#' @param metr.Y see metr.X.
#' @param use specifies how to treat missing values. "complete.obs" excludes NAs, "all" uses all observations.
#' @param return.data logical; specifies if the test object should contain the original data.
#' @param algorithm specifies the algorithm used for calculating the distance covariance. 
#' @param settings a named list of additional tuning parameters, used to override defaults for specific tests. For the spectral test ("spectral") this includes \code{starteig}, \code{multeig}, \code{large}, \code{tolfac}, \code{nomlevel}, \code{maxlz} and \code{weedlimit}; \code{exactmom} controls exact-moment computation. Mainly for advanced or internal use; if \code{NULL}, sensible defaults are chosen.
#' "fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 
#' 
#' "memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 
#' 
#' "standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.
#' 
#' "auto" chooses the best algorithm for the specific setting using a rule of thumb.
#' 
#' @return distcov.test object
#' @export
#' @references
#' \insertRef{berschneider2018complex}{dcortools}
#' 
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{chwialkowski2014wild}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huang2017statistically}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#'
distcov.test <- function(X,
                         Y,
                         method = "permutation",
                         b = 499L,
                         ln = 20,
                         affine = FALSE,
                         standardize = FALSE,
                         bias.corr = FALSE,
                         type.X = "sample",
                         type.Y = "sample",
                         metr.X = "euclidean",
                         metr.Y = "euclidean",
                         use = "all",
                         return.data = FALSE,
                         algorithm = "auto",
                         settings = NULL
                       ) {
  
  smp <- NULL
  #extract dimensions and sample sizes
  ss.dimX <- extract_np(X, type.X)
  ss.dimY <- extract_np(Y, type.Y)
  
  settings$bias.corr <- bias.corr
  
  
  n <- ss.dimX$Sample.Size
  p <- ss.dimX$Dimension
  
  m <- ss.dimY$Sample.Size
  q <- ss.dimY$Dimension
  
  dogamma <- docons  <- doconssz <- dospv <- dobb3 <- dospec <- doperm <- dowild1 <- dowild2 <-  FALSE
  
  
  
  output <- list()
  
  if (method == "gamma")
    dogamma <- TRUE 
  else if (method == "conservativebb")
    docons <- TRUE
  else if (method == "conservativesz")
    doconssz <- TRUE
  else if (method == "spv")
    dospv <- TRUE
  else if (method == "bb3")
    dobb3 <- TRUE
  else if (method == "spectral")
    dospec <- TRUE
  else if (method == "permutation")
    doperm <- TRUE
  else if (method == "wildbs1")
    dowild1 <- TRUE
  else if (method == "wildbs2")
    dowild2 <- TRUE
  else
    stop ("Test must be one of \"permutation\", \"gamma\", \"bb3\", \"spectral\", \"conservativebb\", \"conservativesz\", \"spv\"  \"wildbs1\" or \"wildbs2\" ")
  
  
  if (dospec) {
    if (is.null(settings$starteig)) {
      settings$starteig <- 15
    }
    if (is.null(settings$multeig)) {
      settings$multeig <- 3
    }
    if (is.null(settings$large)) {
      settings$large <- 0.2
    }
    if (is.null(settings$tolfac)) {
      settings$tolfac <- 1.1
    }
    if (is.null(settings$nomlevel)) {
      settings$nomlevel <- 0.05
    }
    if (is.null(settings$maxlz)) {
      settings$maxlz <- floor(0.15 * n)
    }
    if (is.null(settings$weedlimit)) {
      settings$weedlimit <- 1e-6
    }
  }
  if (is.null(settings$exactmom)) {
    settings$exactmom <- TRUE
  }


  
  nlevels <- length(settings$nomlevel)
  
  
  if (n != m) {
    stop("Samples X and Y must have the same sizes!")
  }
  
  if(return.data) {
    output$X <- X
    output$Y <- Y
  } else {
    output <- NULL
  }
  
  
  if (bias.corr == TRUE) {
    termstodcov2 <- function(aijbij, Sab, Tab, n) {
      aijbij / n / (n - 3) - 2 * Sab / n / (n - 2) / (n - 3) + Tab / n / (n - 1) / (n - 2) / (n - 3) 
    }
    dcov2todcov <- function(dcov2) {
      sqrt(abs(dcov2)) * sign(dcov2)
    }
    dcov2todcor <- function(dcov2, dvarX, dvarY) {
      (sqrt(abs(dcov2)) * sign(dcov2)) / sqrt(sqrt(dvarX * dvarY))
    }
  } else  {
    termstodcov2 <- function(aijbij, Sab, Tab, n) {
      aijbij / n / n - 2 * Sab / n / n / n + Tab / n / n / n / n
    }
    dcov2todcov <- function(dcov2) {
      sqrt(dcov2)
    }
    dcov2todcor <- function(dcov2, dvarX, dvarY) {
      sqrt(dcov2) / sqrt(sqrt(dvarX * dvarY))
    }
  }
  
  
  
  if (dogamma) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      if (!settings$exactmom) {
        dvarX <- terms$aijaij / n / (n - 3) - 2 * terms$Saa / n / (n - 2) / (n - 3) + terms$adotdot * terms$adotdot / n / (n - 1) / (n - 2) / (n - 3) 
        dvarY <- terms$bijbij / n / (n - 3) - 2 * terms$Sbb / n / (n - 2) / (n - 3) + terms$bdotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3) 
        est.var <- 2 * dvarX  * dvarY
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 2 / (n - 1) ^ 2
      } else if (settings$exactmom) {
        est.m2 <- sum(moms) / n ^ 10
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 3 / (n - 1)
        est.var <- (est.m2 - est.m1 ^ 2)
      }
      if (settings$bias.corr) {
        stat <- n *  dcov2 + est.m1
      } else {
        stat <- n *  dcov2
      }
      alpha <- est.m1 ^ 2 / est.var
      beta <- est.m1 / est.var
      pval <- stats::pgamma(stat, alpha, beta, lower.tail = FALSE) 
      return(pval)
    }
  }  else if (doperm) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      if (is.na(dcov2))
        return(NA)
      Tab <- terms$adotdot * terms$bdotdot
      reps <- lapply(1:b, function(t) {
        terms.sample <- terms.smp(terms = terms, smp = smp[[t]])
        return(termstodcov2(terms.sample$aijbij, terms.sample$Sab, Tab, n))
      })
      pval <- (1 + length(which(reps >= dcov2))) / (1 + b)
      return(pval)
    }
  } else if (dowild1 | dowild2) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      cX <- doublecent(terms$distX)
      cY <- doublecent(terms$distY)
      dXY <- cX*cY
      stat <- sum(dXY)
      reps <- sapply(1:b, function(t) {
        return(t(smp[[t]]) %*% dXY %*% smp[[t]])
      })
      pval <- (1 + length(which(reps >= stat))) / (1 + b)
      return(pval)
    }
  } else if (docons) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      if (!settings$exactmom) {
        dvarX <- terms$aijaij / n / (n - 3) - 2 * terms$Saa / n / (n - 2) / (n - 3) + terms$adotdot * terms$adotdot / n / (n - 1) / (n - 2) / (n - 3) 
        dvarY <- terms$bijbij / n / (n - 3) - 2 * terms$Sbb / n / (n - 2) / (n - 3) + terms$bdotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3) 
        est.var <- 2 * dvarX  * dvarY
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 2 / (n - 1) ^ 2
      } else if (settings$exactmom) {
        est.m2 <- sum(moms) / n ^ 10
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 3 / (n - 1)
        est.var <- (est.m2 - est.m1 ^ 2)
      }
      if (settings$bias.corr) {
        stat <- n *  dcov2 + est.m1
      } else {
        stat <- n *  dcov2
      }
      alpha <- sqrt(est.var / 2 / est.m1 ^ 2)
      pval <- stats::pchisq(stat * sqrt(2) / sqrt(est.var), df = 1 / alpha, lower.tail = FALSE) 
    }
  } else if (doconssz) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      stat <- terms$aijbij / n - 2 * terms$Sab / n ^ 2 + terms$adotdot * terms$bdotdot / n ^ 3
      statnorm <- stat / (terms$adotdot * terms$bdotdot / n ^ 4) 
      pval <- stats::pchisq(statnorm, df = 1, lower.tail = FALSE) 
    }
  } else if (dospv) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      dcov2 <- terms$aijbij / n / (n - 3) - 2 * terms$Sab / n / (n - 2) / (n - 3) + terms$adotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3)
      dvarX <- terms$aijaij / n / (n - 3) - 2 * terms$Saa / n / (n - 2) / (n - 3) + terms$adotdot * terms$adotdot / n / (n - 1) / (n - 2) / (n - 3) 
      dvarY <- terms$bijbij / n / (n - 3) - 2 * terms$Sbb / n / (n - 2) / (n - 3) + terms$bdotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3) 
      statnorm <- n * (dcov2/sqrt(dvarX * dvarY)) + 1
      pval <- stats::pchisq(statnorm, df = 1, lower.tail = FALSE) 
    }
  } else if (dobb3) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      if (!settings$exactmom) {
        dvarX <- terms$aijaij / n / (n - 3) - 2 * terms$Saa / n / (n - 2) / (n - 3) + terms$adotdot * terms$adotdot / n / (n - 1) / (n - 2) / (n - 3) 
        dvarY <- terms$bijbij / n / (n - 3) - 2 * terms$Sbb / n / (n - 2) / (n - 3) + terms$bdotdot * terms$bdotdot / n / (n - 1) / (n - 2) / (n - 3) 
        est.var <- 2 * dvarX  * dvarY
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 2 / (n - 1) ^ 2
      } else if (settings$exactmom) {
        est.m2 <- sum(moms$vc) / n ^ 10
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 3 / (n - 1)
        est.var <- (est.m2 - est.m1 ^ 2)
      }
      est.skw <- moms$skw
      beta <- est.skw / sqrt(8)
      if (settings$bias.corr) {
        centstat <- n *  dcov2  /  sqrt(est.var)
      } else {
        centstat <- (n *  dcov2 - est.m1) /  sqrt(est.var)
      }
      pval <- stats::pchisq((centstat * sqrt(2) + 1 / beta) / beta , df = 1 / beta ^ 2, lower.tail = FALSE)  
      return(pval)
    } 
  } else if (dospec) {
    testfunc <- function(dcov2, smp, terms, n, moms, settings, ...) {
      if (settings$exactmom) {
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 3 / (n - 1)
        dcX <- doublecent(terms$distX) / sqrt(n * (n-1))
        dcY <- doublecent(terms$distY) / sqrt(n * (n-1))
      } else {
        est.m1 <- terms$adotdot * terms$bdotdot / n ^ 4 
        dcX <- doublecent(terms$distX) / n
        dcY <- doublecent(terms$distY) / n
      }
      if (settings$bias.corr) {
        stat <- n *  dcov2 + est.m1
      } else {
        stat <- n *  dcov2
      }
      stop <- 0
      k <- min(settings$starteig, n)
      
      if (k > settings$maxlz) {stop <- 2}
      
      while (stop < 1) {
        eX <- RSpectra::eigs_sym(-dcX, k = k)$values
        eY <- RSpectra::eigs_sym(-dcY, k = k)$values
        
        
        
        eigscut <- as.vector(outer(eX[1 : k], eY[1 : k]))
        val <- max(eX[1] * eY[k], eY[1] * eX[k])
        rest <- est.m1 - sum(eX[1 : k]) * sum(eY[1 : k])
        
        dfmain <- floor(rest / val)
        rst <- rest/val - dfmain
        
        if (stat > 2*est.m1) {
          acons <- pval_lambda(c(eigscut,val,rst*val), stat, df = c(rep(1, k ^ 2), dfmain, 1))$p.value
        } else {
          acons <- pval_lambda(eigscut, stat)$p.value
        }
        
        if (acons > settings$large) {
          pval <- acons
          stop <- 2
        } else {
          cons <- pval_lambda(c(eigscut,rest), stat, df = rep(1, k ^ 2 + 1))$p.value
          
          if (cons/acons < settings$tolfac | acons == 0) {
            pval <- cons
            stop <- 2
            for (z in 1:nlevels) {
            if ((cons > settings$nomlevel[z]) & (acons < settings$nomlevel[z])) {
              stop <- 0
            }
            }
          }
          k <- min(settings$multeig * k, n)
          if (k == n) {
            stop <- 2
          }
        }
      }
      
      if (k > settings$maxlz) {
        eX <- eigen(-dcX, symmetric = TRUE, only.values = TRUE)$values 
        eY <- eigen(-dcY, symmetric = TRUE, only.values = TRUE)$values
        
        eigsall <- as.vector(outer(eX, eY))
        
        if (settings$exactmom) {
          est.m2 <- sum(moms) / n ^ 10
          est.var <- (est.m2 - est.m1 ^ 2)
          eigsall <- pmax(shrink_to_target_sumsq(eigsall[1:(n-1)^2], est.var/2, (n-1)^2), 0)
        }
        sortev <- sort(eigsall, decreasing = TRUE)
        cs2 <- cumsum(sortev^2)
        small <- max(which(cs2 < (cs2[(n-1)^2] - settings$weedlimit))) + 1
        eigsweed <- sortev[1:small]
        corrterm <- est.m1 - sum(sortev[1:small])
        #pval <- pval_lambda(eigsall, stat)$p.value
        pval <- pval_lambda(eigsweed, stat - corrterm)$p.value
      }
      return(pval)
    } 
  }
  
  
  
  
  
  if (use == "complete.obs") {
    ccX <- ccY <- cc <- 1:n
    if (type.X == "sample") {
      ccX <- which(stats::complete.cases(X))
    }
    if (type.Y == "sample") {
      ccY <- which(stats::complete.cases(Y))
    }
    cc <- intersect(ccX, ccY)
    if (type.X == "sample" && p == 1) {
      X <- X[cc]
    } else if (type.X == "sample" && p > 1) {
      X <- X[cc, ]
    }
    if (type.Y == "sample" && q == 1) {
      Y <- Y[cc]
    } else if (type.X == "sample" && q > 1) {
      Y <- Y[cc, ]
    }
    n <- m <- length(cc)
    
    if (type.X == "distance") {
      X <- X[cc,cc]
    }
    if (type.Y == "distance") {
      Y <- Y[cc,cc]
    }
  } 
  
  
  ## normalize samples if calculation of affinely invariant distance covariance is desired
  if (affine == TRUE) {
    if (type.X == "distance" | type.Y == "distance") {
      stop("Affinely invariant distance covariance cannot be calculated for type 'distance'")
    }
    if (p > n | q > n) {
      stop("Affinely invariant distance covariance cannot be calculated for p>n")
    }
    if (p > 1) {
      X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
    } else {
      X <- X / stats::sd(X)
    }
    if (q > 1) {
      Y <- Y %*% Rfast::spdinv(mroot(stats::var(Y)))
    } else {
      Y <- Y / stats::sd(Y)
    }
  } else if (standardize) {
    if (type.X == "distance" | type.Y == "distance") {
      stop("Standardization cannot be applied for type distance.")
    }
    if (p > 1) {
      X <- standardise(X, center = FALSE)
    } else {
      X <- X / stats::sd(X)
    }
    if (q > 1) {
      Y <- standardise(Y, center = FALSE)
    } else {
      Y <- Y / stats::sd(Y)
    }
  }
  
  if (algorithm == "auto") {
    if (p == 1 & q == 1 & metr.X[1] %in% c("euclidean", "discrete") 
        & metr.Y %in% c("euclidean", "discrete") & n > 100 &  type.X == "sample" & type.Y == "sample" & !(dobb3|dowild1|dowild2|dospec)) {
      algorithm <- "fast"
    } else if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") &
               metr.Y[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") & type.X == "sample" & type.Y == "sample" & !(dobb3|dowild1|dowild2|dospec)) {
      algorithm <- "memsave"
    } else {
      algorithm <- "standard"
    }
  }
  
  alg.fast <- alg.standard <- alg.memsave <- FALSE
  
  if (algorithm == "fast") {
    alg.fast <- TRUE 
    if (doperm) 
      terms.smp <- function(terms, smp) {
        ndisc <- sum(c(metr.X =="discrete",metr.Y=="discrete"))
        if (ndisc == 0) 
          output <- sampleterms.fast(terms, smp)
        else if (ndisc == 1)
          output <- sampleterms.fast.numdisc(terms, smp)
        else if (ndisc == 2) {
          terms$aidot <- terms$aidotshort
          terms$bidot <- terms$bidotshort
          output <- sampleterms.fast.discrete(terms, smp)
        }}
  } else if (algorithm == "standard") {
    alg.standard <- TRUE
    if (doperm) 
      terms.smp <- function(terms, smp) {sampleterms.standard(terms, smp)}
  }  else if (algorithm == "memsave") {
    alg.memsave <- TRUE
    if (doperm) 
      terms.smp <- function(terms, smp) {sampleterms.memsave(terms, smp)}
  } 
  else
    stop ("Algorithm must be one of \"fast\", \"standard\", \"memsave\" or \"auto\"")
  
  
  if (!alg.standard & (dobb3|dowild1|dowild2|dospec))
    stop("bb3 and wild bootstrap p-value calculation is only possible with algorithm=standard!")
  
  
  if (dowild1 | dowild2) {
    perms <- lapply(1:b, function(t) {
      epsilon <- stats::rnorm(n+1)
      W <- rep(NA,n)
      W[1] <- exp(-1/ln) * epsilon[1]+sqrt(1-exp(-2/ln)) *epsilon[2]
      for (r in 2:n) {
        W[r] <- exp(-1/ln) * W[r-1] + sqrt(1-exp(-2/ln))*epsilon[r+1]
      }
      if (dowild2)
        W <- W - mean(W)
      return(W)
    })
  } else if (doperm) {
    if (is.null(settings$blocktype)) {
    perms <- lapply(1:b, function(t) sample(1:n))
    } else if (settings$blocktype == "intra") {
     safe_sample <- function(x) if (length(x) <= 1) x else sample(x)
     perms <- lapply(1:b, function(t) stats::ave(1:n, settings$blocks, FUN = safe_sample))
    }  else if (settings$blocktype == "custom") {
      perms <- settings$perms
    }
  }else {
    perms <- NULL
  }
  
  if (type.X == "distance")
    metr.X <- "distance matrix provided"
  
  if (type.Y == "distance")
    metr.Y <- "distance matrix provided"
  
  terms <- dcovterms(X,Y,n, calc.dcor = TRUE, doperm = doperm, dobb3 = dobb3|dowild1|dowild2|dospec, alg.fast = alg.fast, alg.memsave = alg.memsave, alg.standard = alg.standard, p = p, q = q, metr.X = metr.X, metr.Y =metr.Y, type.X = type.X, type.Y = type.Y)
  
  
  
  
  
  
  dcov2 <- termstodcov2(terms$aijbij, terms$Sab, terms$adotdot * terms$bdotdot, n)
  output$dcov <- dcov2todcov(dcov2)
  dvarX <- termstodcov2(terms$aijaij, terms$Saa, terms$adotdot * terms$adotdot, n)
  dvarY <- termstodcov2(terms$bijbij, terms$Sbb, terms$bdotdot * terms$bdotdot, n)
  output$dsdX <- sqrt(dvarX)
  output$dsdY <- sqrt(dvarY)
  output$dcor <- dcov2todcor(dcov2, dvarX, dvarY)
  
  if (dogamma|docons|dospec) {
    if (settings$exactmom) {
      moms <- calcmom(terms$aijaij, terms$Saa, terms$adotdot, terms$bijbij, terms$Sbb, terms$bdotdot, n = n, dobb3 = FALSE, dospec = dospec)
    }  } else if (dobb3) {
      moms <- calcmom(terms$aijaij, terms$Saa, terms$adotdot, terms$bijbij, terms$Sbb, terms$bdotdot, terms$distX, terms$distY, terms$aidot, terms$bidot,  n, dobb3 = TRUE)
    } else {
      moms <- NULL
    }
  
  
  
  
  output$pvalue <-  testfunc(dcov2 = dcov2, terms = terms, moms = moms, n = n, smp = perms, settings = settings)
  
  class(output) <- "dctest"
  
  output$call <- match.call()
  output$method <- method
  output$affine <- affine
  output$bias.corr <- bias.corr
  output$standardize <- standardize
  output$metr.X <- metr.X
  output$metr.Y <- metr.Y
  output$b <- b
  
  # if (dogamma) 
  #  warning("The simple gamma approximation can be anticonservative, in particular for small p-values.")
  
  return(output)
}




calcmom  <- function(aijaij, Saa, adotdot, bijbij = NULL, Sbb = NULL, bdotdot = NULL, distX = NULL, distY =NULL,  aidot = NULL, bidot =NULL,  n, dobb3, dospec = FALSE) {
  
  n2 <- n ^ 2
  n3 <- n ^ 3
  n4 <- n ^ 4
  
  
  vc.C <- c(n * (n - 1) * (n - 2) * (n - 3),
            2 * n * (n - 1),
            4 * n * (n - 1) * (n - 2),
            n * (n - 1),
            4 * n * (n - 1),
            2 * n * (n - 1) * (n - 2),
            n)
  
  
  vc.b <- c(6 * n + 2 * n2,
            6 * n - 2 * n2 - 2 * n3 + n4,
            6 * n - n3,
            6 * n - 2 * n2,
            6 * n - 4 * n2,
            6 * n,
            6 * n - 10 * n2 + 4 * n3)
  
  vc.c <- c(-24 * n - 4 * n2,
            -24 * n + 12 * n2 + 4 * n3 - 2 * n4,
            -24 * n + 4 * n2 + 2 * n3,
            -24 * n + 12 * n2,
            -24 * n + 20 * n2 - 4 * n3,
            -24 * n + 4 * n2,
            -24 * n + 44 * n2 - 24 * n3 + 4 * n4)
  
  
  vc.d <- c(18 * n + 3 * n2,
            18 * n - 9 * n2 - 2 * n3 + n4,
            18 * n - 3 * n2 -  n3,
            18 * n - 9 * n2 - 2 * n3 + n4,
            18 * n - 15 * n2 + 3 * n3,
            18 * n - 3 * n2 -  n3,
            18 * n - 33 * n2 + 18 * n3 - 3 * n4)
  
  biX <- aijaij / n / (n - 1)
  
  ciX <- (Saa - aijaij) / n / (n - 1) / (n - 2)
  
  diX <- (adotdot ^ 2 + 2 * aijaij - 4 * Saa) / n / (n - 1) / (n - 2) / (n - 3)
  
  vc.X <- vc.b * biX + vc.c * ciX + vc.d * diX
  
  if (!dobb3) {
    if (is.null(bijbij)) {
      return(sqrt(vc.C) * vc.X)
    }  else {
      biY <- bijbij / n / (n - 1)
      ciY <- (Sbb - bijbij) / n / (n - 1) / (n - 2)
      diY <- (bdotdot ^ 2 + 2 * bijbij - 4 * Sbb) / n / (n - 1) / (n - 2) / (n - 3)
      vc.Y <- vc.b * biY + vc.c * ciY + vc.d * diY
      return(vc.C * vc.X * vc.Y)
    }
  } else {
    vc.X.lim <- biX - 2 * ciX + diX
    est.var.X.lim <- sqrt(2) * vc.X.lim
    distXmatp2 <- squaresym(distX)
    amatp2dot <- Rfast::colsums(distXmatp2)
    distX2 <- distX * distX
    a2dot <- Rfast::colsums(distX2)
    
    BiX.C.C <- sum(amatp2dot * aidot)
    BiX.C.H <- matrix_prod_sum(distXmatp2, distX)
    BiX.H.C <- sum(a2dot * aidot)
    BiX.H.H <- sum_hadamard_power3(distX)
    BiX.H <- aijaij
    BiX.C <- Saa
    CS3X <- sum(aidot ^ 3)
    
    eiX <- BiX.C.H / n / (n - 1) / (n - 2)
    fiX <- (BiX.C.C - BiX.C.H - 2 * BiX.H.C + BiX.H.H) / n / (n - 1) / (n - 2) / (n - 3)
    yiX <- (BiX.C * adotdot - BiX.H * adotdot - 2 * CS3X - 4 * BiX.H.H -
              4 * BiX.C.C + 2 * BiX.C.H + 10 * BiX.H.C) /
      n / (n - 1) / (n - 2) / (n - 3) / (n - 4)
    uiX <- (adotdot ^ 3 + 16 * BiX.H.H - 48 * BiX.H.C - 8 * BiX.C.H +
              6 * adotdot * BiX.H + 24 * BiX.C.C + 16 * CS3X - 12 * BiX.C * adotdot) /
      n / (n - 1) / (n - 2) / (n - 3) / (n - 4) / (n - 5)
    
    
    est.m3cent.X <- - 1 * eiX + 3 * fiX - 3 * yiX + uiX
    
    est.m3cent.X <- sqrt(8) * est.m3cent.X
    
    est.skw.X <- est.m3cent.X / est.var.X.lim ^ (3 / 2)
    
    if (!is.na(est.skw.X)) {
      if (est.skw.X < 0)
        est.skw.X <- 1e-3
    }
    
    
    if (is.null(bijbij)) { 
      return(list("vc" = sqrt(vc.C) * vc.X, "skw" = est.skw.X))
    } else {
      
      biY <- bijbij / n / (n - 1)
      ciY <- (Sbb - bijbij) / n / (n - 1) / (n - 2)
      diY <- (bdotdot ^ 2 + 2 * bijbij - 4 * Sbb) / n / (n - 1) / (n - 2) / (n - 3)
      vc.Y <- vc.b * biY + vc.c * ciY + vc.d * diY
      
      vc.Y.lim <- biY - 2 * ciY + diY
      est.var.Y.lim <- sqrt(2) * vc.Y.lim
      distYmatp2 <- squaresym(distY)
      bmatp2dot <- Rfast::colsums(distYmatp2)
      distY2 <- distY * distY
      b2dot <- Rfast::colsums(distY2)
      
      BiY.C.C <- sum(bmatp2dot * bidot)
      BiY.C.H <- matrix_prod_sum(distYmatp2, distY)
      BiY.H.C <- sum(b2dot * bidot)
      BiY.H.H <- sum_hadamard_power3(distY)
      BiY.H <- bijbij
      BiY.C <- Sbb
      CS3Y <- sum(bidot ^ 3)
      
      eiY <- BiY.C.H / n / (n - 1) / (n - 2)
      fiY <- (BiY.C.C - BiY.C.H - 2 * BiY.H.C + BiY.H.H) / n / (n - 1) / (n - 2) / (n - 3)
      yiY <- (BiY.C * bdotdot - BiY.H * bdotdot - 2 * CS3Y - 4 * BiY.H.H -
                4 * BiY.C.C + 2 * BiY.C.H + 10 * BiY.H.C) /
        n / (n - 1) / (n - 2) / (n - 3) / (n - 4)
      uiY <- (bdotdot ^ 3 + 16 * BiY.H.H - 48 * BiY.H.C - 8 * BiY.C.H +
                6 * bdotdot * BiY.H + 24 * BiY.C.C + 16 * CS3Y - 12 * BiY.C * bdotdot) /
        n / (n - 1) / (n - 2) / (n - 3) / (n - 4) / (n - 5)
      
      
      est.m3cent.Y <- - 1 * eiY + 3 * fiY - 3 * yiY + uiY
      
      est.m3cent.Y <- sqrt(8) * est.m3cent.Y
      
      est.skw.Y <- est.m3cent.Y / est.var.Y.lim ^ (3 / 2)
      
      if (!is.na(est.skw.Y)) {
        if (est.skw.Y < 0)
          est.skw.Y <- 1e-3
      }
      
      return(list("vc" = vc.C * vc.X * vc.Y, "skw" = est.skw.X * est.skw.Y, "m3X" = est.m3cent.X *est.m3cent.Y))
    }
  }
}






shrink_to_target_sumsq <- function(eigvals, C, p) {
  mu <- mean(eigvals)
  if (C-p*mu^2 > 0) {
    alpha <- sqrt((C - p * mu ^ 2) / sum((eigvals - mu)^2))
    shrunk_eigvals <- alpha * eigvals + (1 - alpha) * mu
    return(shrunk_eigvals)
  } else {return(eigvals)}
}

pval_lambda <- function(lambda, q, df = rep(1, length(lambda))) {
  dv <- CompQuadForm::davies(q, lambda = lambda, h = df, acc = 1e-6)
  p  <- dv$Qq
  if (dv$ifault != 0 || p <= 0 || p > 1)
    p <- survey::pchisqsum(q, df = df, a = lambda,
                           lower.tail = FALSE, method = "saddlepoint")
  p
}

#' Performs distance correlation sure independence screening \insertCite{li2012feature}{dcortools} with some additional options (such as calculating corresponding tests).
#'
#' @param X A dataframe or matrix.
#' @param Y A vector-valued response having the same length as the number of rows of X.
#' @param k Number of variables that are selected (only used when threshold is not provided).
#' @param threshold If provided, variables with a distance correlation larger than threshold are selected.
#' @param calc.cor If set as "pearson", "spearman" or "kendall", a corresponding correlation matrix is additionally calculated.
#' @param calc.pvalue.cor logical; IF TRUE, a p-value based on the Pearson or Spearman correlation matrix is calculated (not implemented for calc.cor = "kendall") using Hmisc::rcorr.
#' @param return.data logical; specifies if the dcmatrix object should contain the original data.
#' @param test Allows for additionally calculating a test based on distance Covariance. Specifies the type of test that is performed, "permutation" performs a Monte Carlo Permutation test. "gamma" performs a test based on a gamma approximation of the test statistic under the null. "conservative" performs a conservative two-moment approximation. "bb3" performs a quite precise three-moment approximation and is recommended when computation time is not an issue.
#' @param adjustp If setting this parameter to "holm", "hochberg", "hommel", "bonferroni", "BH", "BY" or "fdr", corresponding adjusted p-values are additionally returned for the distance covariance test.
#' @param b specifies the number of random permutations used for the permutation test. Ignored for all other tests.
#' @param bias.corr logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.
#' @param use "all" uses all observations, "complete.obs" excludes NAs, "pairwise.complete.obs" uses pairwise complete observations for each comparison.
#' @param algorithm specifies the algorithm used for calculating the distance covariance. 
#' 
#' "fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 
#' 
#' "memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 
#' 
#' "standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.
#' 
#' "auto" chooses the best algorithm for the specific setting using a rule of thumb.
#' 
#' "memsave" is typically very inefficient for dcsis and should only be applied in exceptional cases.
#' @return dcmatrix object with the following two additional slots:
#' \item{name selected}{description indices of selected variables.} 
#' \item{name dcor.selected}{distance correlation of the selected variables and the response Y.}
#' @export
#' @references
#' \insertRef{berschneider2018complex}{dcortools}

#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huang2017statistically}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{li2012feature}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @examples
#' X <- matrix(rnorm(1e5), ncol = 1000)
#' Y <- sapply(1:100, function(u) sum(X[u, 1:50])) + rnorm(100)
#' a <- dcsis(X, Y)

dcsis <- function(X, 
                  Y,
                  k = floor(nrow(X)/log(nrow(X))),
                  threshold = NULL,
                  calc.cor = "spearman",
                  calc.pvalue.cor = FALSE,
                  return.data = FALSE,
                  test = "none",
                  adjustp = "none",
                  b = 499,
                  bias.corr = FALSE,
                  use="all",
                  algorithm ="auto") {
  
  p <- ncol(X)
  q <- ncol(Y)
  
  if (is.null(q))
    q <- 1
  
  if (k  < 0) {
    stop("k must be positive")
  }
  
  if (q != 1) 
    stop("Dimension of response must be 1 for screening algorithm!")
  
  output <- dcmatrix(X = X,
                     Y = Y,
                     calc.cor = calc.cor,
                     calc.pvalue.cor = calc.pvalue.cor,
                     return.data = return.data,
                     test = test,
                     adjustp = adjustp,
                     b = b,
                     bias.corr = bias.corr,
                     use = use,
                     algorithm = algorithm)
  
  if (!is.null(threshold)) {
    output$selected <- which(calc.cor > threshold)
    output$dcor.selected <- output$dcor[output$selected]
  } else {
    k <- min(round(k),p)
    
    output$selected <- order(output$dcor, decreasing = TRUE)[1:k]
    output$dcor.selected <- output$dcor[output$selected]
  }
  
  return(output)  

  
}

#' Calculates an inverse-probability-of-censoring weighted (IPCW) distance correlation based on IPCW U-statistics \insertCite{datta2010inverse}{dcortools}.
#'
#' @param Y A matrix with two columns, where the first column contains the survival times and the second column the status indicators (a survival object will work).
#' @param X A vector or matrix containing the covariate information.
#' @param affine logical; specifies if X should be transformed such that the result is invariant under affine transformations of X
#' @param standardize logical; should X be standardized using the standard deviations of single observations?. No effect when affine = TRUE.
#' @param timetrafo specifies a transformation applied on the follow-up times. Can be "none", "log" or a user-specified function.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.
#' @param metr.X specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").
#' 
#'  Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
#'  
#'  For "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric,parameter)", c("gaussian",3) for example uses a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
#' @param use specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.
#' @param cutoff If provided, all survival times larger than cutoff are set to the cutoff and all corresponding status indicators are set to one. Under most circumstances, choosing a cutoff is highly recommended.
#' @return An inverse-probability of censoring weighted estimate for the distance correlation between X and the survival times.
#' @export
#' @references
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{datta2010inverse}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @examples 
#' X <- rnorm(100)
#' survtime <- rgamma(100, abs(X))
#' cens <- rexp(100)
#' status <- as.numeric(survtime < cens)
#' time <- sapply(1:100, function(u) min(survtime[u], cens[u]))
#' surv <- cbind(time, status)
#' ipcw.dcor(surv, X)

ipcw.dcor <- function(Y, X, affine = FALSE, standardize = FALSE, timetrafo = "none", type.X = "sample", metr.X = "euclidean", use = "all", cutoff = NULL) {
  
  dcor <- ipcw.dcov(Y, X, affine, standardize, timetrafo, type.X, metr.X, use, cutoff) / sqrt(ipcw.dcov(Y, Y[,1], affine, standardize, timetrafo, type.X = "sample", metr.X, use, cutoff)) / sqrt(distsd(X =  X, affine = affine, standardize = standardize, type.X = type.X, metr.X = metr.X, use = use, bias.corr=TRUE))
  
  return(dcor)
}



#' Calculates an inverse-probability-of-censoring weighted (IPCW) distance covariance based on IPCW U-statistics \insertCite{datta2010inverse}{dcortools}.
#'
#' @param Y A column with two rows, where the first row contains the survival times and the second row the status indicators (a survival object will work).
#' @param X A vector or matrix containing the covariate information.
#' @param affine logical; indicates if X should be transformed such that the result is invariant under affine transformations of X
#' @param standardize logical; should X be standardized using the standard deviations of single observations?. No effect when affine = TRUE.
#' @param timetrafo specifies a transformation applied on the follow-up times. Can be "none", "log" or a user-specified function.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample" (or any other value), X is interpreted as a sample
#' @param metr.X metr.X specifies the metric which should be used for X to analyze the distance covariance. Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq". For "alpha", "minkowski", "gauss", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric,parameter)" (see examples); the standard parameter is 2 for "minkowski" and "1" for all other metrics.
#' @param use specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.
#' @param cutoff If provided, all survival times larger than cutoff are set to the cutoff and all corresponding status indicators are set to one. Under most circumstances, choosing a cutoff is highly recommended.
#' @return An inverse-probability of censoring weighted estimate for the distance covariance between X and the survival times.
#' @export
#' @references
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{datta2010inverse}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @examples 
#' X <- rnorm(100)
#' survtime <- rgamma(100, abs(X))
#' cens <- rexp(100)
#' status <- as.numeric(survtime < cens)
#' time <- sapply(1:100, function(u) min(survtime[u], cens[u]))
#' surv <- cbind(time, status)
#' ipcw.dcov(surv, X)

ipcw.dcov <- function(Y, X, affine = FALSE, standardize = FALSE,  timetrafo = "none", type.X = "sample", metr.X = "euclidean", use = "all", cutoff = NULL) {

    # extract sample size and dimension

    ss.dimX <- extract_np(X, "sample")

    n <- ss.dimX$Sample.Size
    p <- ss.dimX$Dimension
    


    # sort data in order of follow-up times
    
    if (ncol(Y) != 2)
      stop("Y must be a matrix with two columns with time in the first column and status indicator in the second.")
    
    
    if (use == "complete.obs") {
      ccY <- stats::complete.cases(Y)
      Y <- Y[ccY,]
      n <- length(ccY)
      if (type.X == "sample" && p == 1) {
        X <- X[ccY]
      } else if (type.X == "sample" && p > 1) {
        X <- X[ccY, ]
      }
      if (type.X == "distance") {
        X <- X[ccY,ccY]
      }
    }
    
    time <- Y[,1]
    status <- Y[,2]

    IX <- Rfast::Order(time)

    time <- time[IX]
    status <- status[IX]
    
    if (timetrafo != "none") {
      if (timetrafo == "log")
        time <- log(time)
        else {
        trafo <- match.fun(timetrafo)
        time <- trafo(time)
        }
    }

    
    if (type.X == "sample" && p == 1) {
      X <- X[IX]
    } else if (type.X == "sample" && p > 1) {
      X <- X[IX, ]
    }
    if (type.X == "distance") {
      X <- X[IX,IX]
    }
    

    
    if (use == "complete.obs") {
      ccX <- which(stats::complete.cases(X))
      n <- length(ccX)
      if (type.X == "sample" && p == 1) {
        X <- X[ccX]
      } else if (type.X == "sample" && p > 1) {
        X <- X[ccX, ]
      }
      if (type.X == "distance") {
        X <- X[ccX,ccX]
      }
      time <- time[ccX]
      status <- status[ccX]
    }
    
    
    
    
    if (!is.null(cutoff)) {
      time[time>=cutoff] <- cutoff
      status[time>=cutoff] <- 1
    }
    
    events <- which(status == 1)
    
    # calculate IPC weights
    
    
    ipcw <- rep(0, n)
    ipcw[events[1]] <- 1 / (n - events[1] + 1)
    
    help <- sapply(1:n, function(x) ((n - x) / (n - x + 1)) ^ status[x])
    
    for (i in (events[1] + 1):n) {
      ipcw[i]<- prod(help[1:(i - 1)]) * (status[i] / (n - i + 1))
    }
    
    ipcw <- ipcw * n
    
    ipcw <- ipcw[events]
    time <- time[events]
    status <- status[events]
    
    if (affine == TRUE) {
      if (type.X == "distance") {
        stop("Affinely invariant distance variance cannot be calculated for type distance")
      }
      if (p > n) {
        stop("Affinely invariant distance variance cannot be calculated for p>n")
      }
      if (p > 1) {
        X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
      } else {
        X <- X / stats::sd(X)
      }
    } else if (standardize) {
      if (type.X == "distance") {
        stop("Standardization cannot be applied for type distance.")
      }
      if (p > 1) {
        X <- standardise(X, center = FALSE)
      } else {
        X <- X / stats::sd(X)
      }
    }
    


    

    #distance matrices

    if (type.X == "distance") {
      distXall <- X
    } else {
      distXall <- distmat(X, metr.X, p)
    }
    distX <- distXall[events, events]
    
    distT <- Rfast::Dist(time)
    k <- sum(status)

    # auxiliary terms

    m <- sum(ipcw)
    M <- sum(ipcw^2)
    W2mat <- ipcw %*% t(ipcw)
    ipcwm <- matrix(rep(ipcw, k), ncol=k)
    Wmatsum <- ipcwm + t(ipcwm)

    rowsumsX_w <- Rfast::colsums(ipcw * distX)
    rowsumsT_w <- Rfast::colsums(ipcw * distT)
    rowsumsX_w2 <- Rfast::colsums(ipcw^2 * distX)
    rowsumsT_w2 <- Rfast::colsums(ipcw^2 * distT)

    #summands of squared dcov

    term1 <- sum(W2mat * distX * distT * (m ^ 2 -m * Wmatsum - M + 2 * W2mat))
    term2 <- sum(ipcw * (m + ipcw) * rowsumsX_w * rowsumsT_w)
    term3 <- sum(ipcw * rowsumsX_w2 * rowsumsT_w)
    term4 <- sum(ipcw * rowsumsX_w * rowsumsT_w2)
    term5 <- sum(W2mat * distX) * sum(W2mat * distT)

    #calculate dcov and return

    dcov2 <- 1 / (n * (n-1) * (n-2) * (n-3)) * (term1 - 2 * (term2 - term3 - term4) + term5)

    dcov <- sign(dcov2) * sqrt(abs(dcov2))
    return(dcov)
}



#' Performs a permutation test based on the IPCW distance covariance.
#'
#' @param Y A column with two rows, where the first row contains the survival times and the second row the status indicators (a survival object will work).
#' @param X A vector or matrix containing the covariate information.
#' @param affine logical; indicates if X should be transformed such that the result is invariant under affine transformations of X.
#' @param standardize logical; should X be standardized using the standard deviations of single observations. No effect when affine = TRUE.
#' @param timetrafo specifies a transformation applied on the follow-up times. Can be "none", "log" or a user-specified function.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample" (or any other value), X is interpreted as a sample.
#' @param metr.X metr.X specifies the metric which should be used for X to analyze the distance covariance. Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq". For "alpha", "minkowski", "gauss", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric,parameter)" (see examples); the standard parameter is 2 for "minkowski" and "1" for all other metrics.
#' @param use specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.
#' @param cutoff If provided, all survival times larger than cutoff are set to the cutoff and all corresponding status indicators are set to one. Under most circumstances, choosing a cutoff is highly recommended.
#' @param B The number of permutations used for the permutation test
#' @return An list with two arguments, $dcov contains the IPCW distance covariance, $pvalue the corresponding p-value
#' @export
#' @references
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{datta2010inverse}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @examples 
#' X <- rnorm(100)
#' survtime <- rgamma(100, abs(X))
#' cens <- rexp(100)
#' status <- as.numeric(survtime < cens)
#' time <- sapply(1:100, function(u) min(survtime[u], cens[u]))
#' surv <- cbind(time, status)
#' ipcw.dcov.test(surv, X)
#' ipcw.dcov.test(surv, X, cutoff = quantile(time, 0.8)) 
#' # often better performance when using a cutoff time
#' 
ipcw.dcov.test <- function(Y, X, affine = FALSE, standardize = FALSE, timetrafo = "none", type.X = "sample", metr.X = "euclidean", use = "all", cutoff = NULL, B=499)
{
    ss.dimX <- extract_np(X, "sample")

    n <- ss.dimX$Sample.Size
    p <- ss.dimX$Dimension
    
    if (ncol(Y) != 2)
      stop("Y must be a matrix with two columns with time in the first column and status indicator in the second.")
    
    
    if (use == "complete.obs") {
      ccY <- stats::complete.cases(Y)
      Y <- Y[ccY,]
      n <- length(ccY)
      if (type.X == "sample" && p == 1) {
        X <- X[ccY]
      } else if (type.X == "sample" && p > 1) {
        X <- X[ccY, ]
      }
      if (type.X == "distance") {
        X <- X[ccY,ccY]
      }
    }
    
    time <- Y[,1]
    status <- Y[,2]
    
   
      

    IX <- Rfast::Order(time)

    time <- time[IX]
    status <- status[IX]
    
    
    if (timetrafo != "none") {
      if (timetrafo == "log")
        time <- log(time)
      else {
        trafo <- match.fun(timetrafo)
        time <- trafo(time)
      }
    }

    if (type.X == "sample" && p == 1) {
      X <- X[IX]
    } else if (type.X == "sample" && p > 1) {
      X <- X[IX, ]
    }
    if (type.X == "distance") {
      X <- X[IX,IX]
    }
    


 

    if (use == "complete.obs") {
      ccX <- which(stats::complete.cases(X))
      n <- length(ccX)
    if (type.X == "sample" && p == 1) {
        X <- X[ccX]
      } else if (type.X == "sample" && p > 1) {
        X <- X[ccX, ]
      }
      if (type.X == "distance") {
        X <- X[ccX,ccX]
      }
      time <- time[ccX]
      status <- status[ccX]
    }
    
    
    
    if (!is.null(cutoff)) {
      time[time>=cutoff] <- cutoff
      status[time>=cutoff] <- 1
    }
    
    events <- which(status == 1)
    # calculate IPC weights
    
    
    ipcw <- rep(0, n)
    ipcw[events[1]] <- 1 / (n - events[1] + 1)
    
    help <- sapply(1:n, function(x) ((n - x) / (n - x + 1)) ^ status[x])
    
    for (i in (events[1] + 1):n) {
      ipcw[i]<- prod(help[1:(i - 1)]) * (status[i] / (n - i + 1))
    }
    
    ipcw <- ipcw * n
    
    ipcw <- ipcw[events]
    time <- time[events]
    status <- status[events]
    
    
    
    
    if (affine == TRUE) {
      if (type.X == "distance") {
        stop("Affinely invariant distance variance cannot be calculated for type distance")
      }
      if (p > n) {
        stop("Affinely invariant distance variance cannot be calculated for p>n")
      }
      if (p > 1) {
        X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
      } else {
        X <- X / stats::sd(X)
      }
    } else if (standardize) {
      if (type.X == "distance") {
        stop("Standardization cannot be applied for type distance.")
      }
      if (p > 1) {
        X <- standardise(X, center = FALSE)
      } else {
        X <- X / stats::sd(X)
      }
    }

    if (type.X == "distance") {
      distXall <- X
    } else {
      distXall <- distmat(X, metr.X, p)
    }
    distX <- distXall[events, events]
    distT <- Rfast::Dist(time)
    k <- sum(status)

    m <- sum(ipcw)
    M <- sum(ipcw^2)
    W2mat <- ipcw%*%t(ipcw)
    Wmatsum <- matrix(rep(ipcw, k), ncol=k) + matrix(rep(ipcw, k), ncol=k, byrow=TRUE)

    term1 <- sum(W2mat * distX * distT * (m ^ 2 -m * Wmatsum - M + 2 * W2mat))

    rowsumsX_w <- Rfast::colsums(ipcw*distX)
    rowsumsT_w <- Rfast::colsums(ipcw*distT)
    rowsumsX_w2 <- Rfast::colsums(ipcw^2*distX)
    rowsumsT_w2 <- Rfast::colsums(ipcw^2*distT)
    term2 <- sum(ipcw * (m + ipcw) * rowsumsX_w * rowsumsT_w)
    term3 <- sum(ipcw * rowsumsX_w2 * rowsumsT_w)
    term4 <- sum(ipcw * rowsumsX_w * rowsumsT_w2)
    term5 <- sum(W2mat * distX) * sum(W2mat * distT)

    dcov2 <- 1 / (n * (n-1) * (n-2) * (n-3)) * (term1 - 2 * (term2 - term3 - term4) + term5)

    dcov <- sign(dcov2) * sqrt(abs(dcov2))




    samp <- lapply(1:B, function(x) sample(1:n))

    reps <- sapply(1:B,
                 function(x) {
                   events_samp <- samp[[x]][events]
                   distX_samp <- distXall[events_samp,events_samp]

                   rowsumsX_w <- Rfast::colsums(ipcw*distX_samp)
                   rowsumsX_w2 <- Rfast::colsums(ipcw^2*distX_samp)


                   term1 <- sum(W2mat * distX_samp * distT * (m ^ 2 -m * Wmatsum - M + 2 * W2mat))


                   term2 <- sum(ipcw * (m + ipcw) * rowsumsX_w * rowsumsT_w)
                   term3 <- sum(ipcw * rowsumsX_w2 * rowsumsT_w)
                   term4 <- sum(ipcw * rowsumsX_w * rowsumsT_w2)
                   term5 <- sum(W2mat * distX_samp) * sum(W2mat * distT)

                   res2 <- 1 / (n * (n-1) * (n-2) * (n-3)) * (term1 - 2 * (term2 - term3 - term4) + term5)
                   res <- sign(res2) * sqrt(abs(res2))
                   return(res)
                 })

  pval <- (1 + length(which(reps>dcov))) / (1 + B)


  return(list("dcov"=dcov, "pvalue"=pval))
}

prep.fast <- function(X, n, discrete, pairwise) {
  
  cc <- mc <- NULL
  ncc <- n
  
  if (!discrete) {
    IX <- temp <- 1:n
   
    
    if (pairwise) {
      IX0 <- NULL
      cc <- which(stats::complete.cases(X))
      mc <- setdiff(1:n,cc)
      
      ncc <- length(cc)
      IX0cc <- Rfast::Order(X[cc])
      IX[cc][IX0cc] <- 1:ncc
      IX[mc] <- NA
      
      IXcc <- IX[cc]
      Xcc <- X[cc]
      X.sort.X <- Xcc[IX0cc]
    } else {
      IX0 <- Rfast::Order(X)
      X.sort.X <- X[IX0]
      IX[IX0] <- temp
      IXcc <- IX
      Xcc <- X
    }
      
    
    
    sX.X <- cumsum(X.sort.X)
    alphaX <- IXcc - 1
    betaX <- sX.X[IXcc] - X.sort.X[IXcc]
    Xdot <- sX.X[ncc]
    aidot <- Xdot + (2 * alphaX - ncc) * Xcc - 2 * betaX
    adotdot <- sum(aidot)
    
    XXdot <- sum(Xcc^2)
    aijaij <- 2 * ncc * XXdot - 2* Xdot^2
    
    return(list("cc" =cc, "mc" = mc, "IX"= IX, "IX0" = IX0, "X.sort.X" = X.sort.X, "aidot" = aidot, "adotdot" = adotdot, "aijaij" = aijaij, "ncc" = ncc, "sX.X" = sX.X, "X" = X))
  } else {
    Xcc <- X
    if (pairwise) {
      cc <- which(stats::complete.cases(X))
      n <- length(cc)
      Xcc <- X[cc]
    }
    nX <- table(Xcc)
    aijaij <- n * (n - 1) - sum(nX * (nX - 1)) 
    aidots <- as.numeric(n - nX)
    adotdot <- sum(aidots * nX)
    aidot <- as.numeric(n - nX[Xcc])
    
    return(list("cc" =cc, "aidotshort" = aidots, "aidot" = aidot, "adotdot" = adotdot, "aijaij" = aijaij, "X" = as.factor(X), "nX" = nX, "ncc" = n))
  }
}





prep.standard <- function(X, n, p,  metr.X, pairwise) {
  
  distX <- distmat(X, metr.X, p)
  
  if (!pairwise) {
    aidot <- Rfast::colsums(distX)
    adotdot <- sum(aidot)
    aijaij <- matrix_prod_sum(distX, distX)
    
    return(list("distX" = distX, "aidot" = aidot, "adotdot" = adotdot, "aijaij" = aijaij, "ncc" = n))
    
  } else {
    cc <- which(stats::complete.cases(X))
    ncc <- length(cc)
    aidot <- colsums_subset(distX,cc)
    adotdot <- sum(aidot)
    aijaij <- matrix_prod_sum_subset(distX, distX, cc)
    return(list("distX" = distX, "aidot" = aidot, "adotdot" = adotdot, "aijaij" = aijaij,  "ncc" = ncc, "cc" = cc))
  }
  
}

prep.memsave <- function(X, n, p,  metr.X, pairwise) {
  if (pairwise) {
    cc <- which(stats::complete.cases(X))
    Xcc <- X[cc]
    n <- length(cc)
  } else {
    Xcc <- X
  }
  
  output <- dvarterms.memsave(Xcc, metr.X, p)
  
  output$ncc <- n
  output$p <- p
  output$X <- X
  
  if (pairwise) {
    output$cc <- cc
  }
  
  
  return(output)
  
}
  

#' @useDynLib dcortools, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom Rfast standardise
NULL

#' @importFrom Rdpack reprompt
NULL

#' Calculates the distance covariance \insertCite{szekely2007,szekely2009brownian}{dcortools}.
#'
#' @param X contains either the first  sample or its corresponding distance matrix.
#'
#' In the first case, X can be provided either as a vector (if one-dimensional), a matrix or a data.frame (if two-dimensional or higher). 
#'
#' In the second case, the input must be a distance matrix corresponding to the sample of interest.
#'
#' If X is a sample, type.X must be specified as "sample". If X is a distance matrix, type.X must be specified as "distance".
#' @param Y see X.
#' @param affine logical; specifies if the affinely invariant distance covariance \insertCite{dueck2014affinely}{dcortools} should be calculated or not.
#' @param standardize logical; specifies if X and Y should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.
#' @param bias.corr logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.
#' @param type.Y see type.X.
#' @param metr.X specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").
#' 
#'  Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
#'  
#'  For "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric, parameter)", e.g. c("gaussian", 3) for a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
#'  
#'  See \insertCite{lyons2013distance,sejdinovic2013equivalence,bottcher2017detecting;textual}{dcortools} for details.
#' @param metr.Y see metr.X.
#' @param use specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.
#' @param algorithm specifies the algorithm used for calculating the distance covariance. 
#' 
#' "fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 
#' 
#' "memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 
#' 
#' "standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.
#' 
#' "auto" chooses the best algorithm for the specific setting using a rule of thumb.
#' 
#' @return numeric; the distance covariance between samples X and Y.
#' @references
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @examples 
#' X <- rnorm(100)
#' Y <- X + 3 * rnorm(100)
#' distcov(X, Y) # standard distance covariance
#' 
#' distcov(X, Y, metr.X = "gaussauto", metr.Y = "gaussauto") 
#' # Gaussian distance with bandwidth choice based on median heuristic
#' 
#' distcov(X, Y, metr.X = c("alpha", 0.5), metr.Y = c("alpha", 0.5)) 
#' # alpha distance covariance with alpha = 0.5.
#' 
#' 
#' #Define a user-specified (slow) version of the alpha metric
#' 
#' alpha_user <- function(X, prm = 1, kernel = FALSE) {
#'     as.matrix(dist(X)) ^ prm
#' }
#' 
#' distcov(X, Y, metr.X = c("alpha", 0.5), metr.Y = c("alpha", 0.5)) 
#' # Gives the same result as before.
#'    
#'
#' #User-specified Gaussian kernel function  
#'      
#' gauss_kernel <- function(X, prm = 1, kernel = TRUE)  {
#'     exp(as.matrix(dist(X)) ^ 2 / 2 / prm ^ 2)
#' }  
#' 
#' distcov(X, Y, metr.X = c("gauss_kernel", 2), metr.Y = c("gauss_kernel", 2)) 
#' # calculates the distance covariance using the corresponding kernel-induced metric
#' 
#' distcov(X, Y, metr.X = c("gaussian", 2), metr.Y = c("gaussian", 2)) 
#' # same result
#' 
#' Y <- matrix(nrow = 100, ncol = 2)
#' X <- rnorm(300)
#' dim(X) <- c(100, 3)
#' Z <- rnorm(100)
#' Y <- matrix(nrow = 100, ncol = 2)
#' Y[, 1] <- X[, 1] + Z
#' Y[, 2] <- 3 * Z
#' 
#' distcov(X, Y) 
#' 
#' distcov(X, Y, affine = TRUE) 
#' # affinely invariant distance covariance
#' 
#' distcov(X, Y, standardize = TRUE) 
#' ## distance covariance standardizing the components of X and Y
#' 
#' @export
distcov <-
  function(X,
           Y,
           affine = FALSE,
           standardize=FALSE,
           bias.corr = FALSE,
           type.X = "sample",
           type.Y = "sample",
           metr.X = "euclidean",
           metr.Y = "euclidean",
           use = "all",
           algorithm = "auto") {
   
    #extract dimensions and sample sizes
    ss.dimX <- extract_np(X, type.X)
    ss.dimY <- extract_np(Y, type.Y)
    
    n <- ss.dimX$Sample.Size
    p <- ss.dimX$Dimension
    
    m <- ss.dimY$Sample.Size
    q <- ss.dimY$Dimension
    
    if (n != m) {
      stop("Samples X and Y must have the same sizes!")
    }
    
    if  (use == "complete.obs") {
      ccX <- ccY <- cc <- 1:n
      ccX <- which(stats::complete.cases(X))
      ccY <- which(stats::complete.cases(Y))
      cc <- intersect(ccX, ccY)
      n <- m <- length(cc)
      if (type.X == "sample" && p == 1) {
        X <- X[cc]
      } else if (type.X == "sample" && p > 1) {
        X <- X[cc, ]
      }
      if (type.Y == "sample" && q == 1) {
        Y <- Y[cc]
      } else if (type.X == "sample" && q > 1) {
        Y <- Y[cc, ]
      }
 
      
      if (type.X == "distance") {
        X <- X[cc,cc]
      }
      if (type.Y == "distance") {
        Y <- Y[cc,cc]
      }
    }
    
    ## normalize samples if calculation of affinely invariant distance covariance is desired
    if (affine) {
      if (type.X == "distance" | type.Y == "distance") {
        stop("Affinely invariant distance covariance cannot be calculated for type distance")
      }
      if (p > n | q > n) {
        stop("Affinely invariant distance covariance cannot be calculated for p>n")
      }
      if (p > 1) {
        X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
      } else {
        X <- X / stats::sd(X)
      }
      if (q > 1) {
        Y <- Y %*% Rfast::spdinv(mroot(stats::var(Y)))
      } else {
        Y <- Y / stats::sd(Y)
      }
    } else if (standardize) {
      if (type.X == "distance" | type.Y == "distance") {
        stop("Standardization cannot be applied for type distance.")
      }
      if (p > 1) {
        X <- standardise(X, center = FALSE)
      } else {
        X <- X / stats::sd(X)
      }
      if (q > 1) {
        Y <- standardise(Y, center = FALSE)
      } else {
        Y <- Y / stats::sd(Y)
      }
    }
    
    
    alg.fast <- alg.standard <- alg.memsave <- FALSE
    
    if (algorithm == "auto") {
      if (p == 1 & q == 1 & metr.X[1] %in% c("euclidean", "discrete") 
          & metr.Y[1] %in% c("euclidean", "discrete") & type.X == "sample" & type.Y == "sample" & n > 100) {
        alg.fast <- TRUE
      } else if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") &
                 metr.Y[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") & type.X == "sample" & type.Y == "sample") {
        alg.memsave <- TRUE
      } else {
        alg.standard <- TRUE
      }
    } else if (algorithm == "fast") {
      alg.fast <- TRUE 
    } else if (algorithm == "standard") {
      alg.standard <- TRUE
    }  else if (algorithm == "memsave") {
      alg.memsave <- TRUE
    } else
      stop ("Algorithm must be one of \"fast\", \"standard\", \"memsave\" or \"auto\"")
    

    terms <- dcovterms(X, Y, n, calc.dcor = FALSE, doperm = FALSE, dobb3 = FALSE, alg.fast = alg.fast, alg.memsave = alg.memsave, alg.standard = alg.standard, p = p, q = q, metr.X = metr.X, metr.Y =metr.Y, type.X = type.X, type.Y = type.Y)

   
    if (bias.corr) {
      dcov2 <- terms$aijbij / n / (n - 3) - 2 * terms$Sab / n / (n - 2) / (n - 3) + terms$Tab / n / (n - 1) / (n - 2) / (n - 3)
      dcov <- sqrt(abs(dcov2)) * sign(dcov2) 
    } else {
      dcov2 <- terms$aijbij / n / n - 2 * terms$Sab / n / n / n  + terms$Tab / n / n / n / n
      dcov <- sqrt(abs(dcov2))
    }
    return(dcov)
}




#' Calculates the distance standard deviation \insertCite{edelmann2017distance}{dcortools}.
#'
#' @param X contains either the sample or its corresponding distance matrix.
#'
#' In the first case, X can be provided either as a vector (if one-dimensional), a matrix or a data.frame (if two-dimensional or higher). 
#'
#' In the second case, the input must be a distance matrix corresponding to the sample of interest.
#'
#' If X is a sample, type.X must be specified as "sample". If X is a distance matrix, type.X must be specified as "distance".
#' @param affine logical; specifies if the affinely invariant distance standard deviation \insertCite{dueck2014affinely}{dcortools} should be calculated or not.
#' @param standardize logical; specifies if X and Y should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.
#' @param bias.corr logical; specifies if the bias corrected version of the sample distance standard deviation \insertCite{huo2016fast}{dcortools} should be calculated.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.
#' @param metr.X specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").
#' 
#'  Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
#'  
#'  For "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric, parameter)", e.g. c("gaussian", 3) for a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
#'  
#'  See \insertCite{lyons2013distance,sejdinovic2013equivalence,bottcher2017detecting;textual}{dcortools} for details.
#' @param use specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.
#' @param algorithm specifies the algorithm used for calculating the distance standard deviation. 
#' 
#' "fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 
#' 
#' "memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 
#' 
#' "standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.
#' 
#' "auto" chooses the best algorithm for the specific setting using a rule of thumb.
#' 
#' @return numeric; the distance standard deviation of X.
#' @references
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{edelmann2017distance}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @export
#' @examples 
#' X <- rnorm(100)
#' distsd(X) # for more examples on the options see the documentation of distcov.
distsd <-
  function(X,
           affine = FALSE,
           standardize=FALSE,
           bias.corr = FALSE,
           type.X = "sample",
           metr.X = "euclidean",
           use = "all",
           algorithm = "auto") {
    #extract dimensions and sample sizes
    ss.dimX <- extract_np(X, type.X)
    
    n <- ss.dimX$Sample.Size
    p <- ss.dimX$Dimension
    
    if (use == "complete.obs") {
      ccX <-  1:n
      n <- length(ccX)
      if (type.X == "sample") {
        ccX <- which(stats::complete.cases(X))
      }
      if (type.X == "sample" && p == 1) {
        X <- X[ccX]
      } else if (type.X == "sample" && p > 1) {
        X <- X[ccX, ]
      }
      if (type.X == "distance") {
        X <- X[ccX,ccX]
      }
    }
    
    
    ## normalize samples if calculation of affinely invariant distance covariance is desired
    if (affine == TRUE) {
      if (type.X == "distance") {
        stop("Affinely invariant distance variance cannot be calculated for type distance")
      }
      if (p > n) {
        stop("Affinely invariant distance variance cannot be calculated for p>n")
      }
      if (p > 1) {
        X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
      } else {
        X <- X / stats::sd(X)
      }
    } else if (standardize) {
      if (type.X == "distance") {
        stop("Standardization cannot be applied for type distance.")
      }
      if (p > 1) {
        X <- standardise(X, center = FALSE)
      } else {
        X <- X / stats::sd(X)
      }
    }
    
    
    alg.fast <- alg.standard <- alg.memsave <- FALSE
    
    if (algorithm == "auto") {
      if (p == 1 & metr.X[1] %in% c("euclidean", "discrete") 
            & type.X == "sample" & n > 100) {
        alg.fast <- TRUE
      } else if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") & type.X == "sample") {
        alg.memsave <- TRUE
      } else {
        alg.standard <- TRUE
      }
    } else if (algorithm == "fast") {
      alg.fast <- TRUE 
    } else if (algorithm == "standard") {
      alg.standard <- TRUE
    }  else if (algorithm == "memsave") {
      alg.memsave <- TRUE
    } else
      stop ("Algorithm must be one of \"fast\", \"standard\", \"memsave\" or \"auto\"")
    
    if (alg.fast) {
      if (p == 1) {
        if (metr.X[1] == "euclidean") {
          terms <- dcovterms.fast(X = X, n = n, calc.dvar = TRUE)
        } else if (metr.X[1] == "discrete") {
          X <- as.factor(X)
          terms <- dcovterms.fast.discrete(X = X, n = n, calc.dvar =TRUE)
        } else {
          stop("metr.X has to be \"euclidean\" or \"discrete\" for fast algorithms")
        }
      } else {
        stop("Dimensions of X must be 1 for fast algorithms.")
      }
    } else if (alg.memsave) {
      if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete")) {
        terms <- dvarterms.memsave(X, metr.X, p)
      } else {
        stop("Memory efficient algorithms cannot be run with user-defined metrics")
      }
    } else if (alg.standard) {
      terms <- dcovterms.standard(X = X, type.X = type.X, metr.X = metr.X, p = p, calc.dvar = TRUE)
    }
    
    
    if (bias.corr) {
      dvar <- terms$aijaij / n / (n - 3) - 2 * terms$Saa / n / (n - 2) / (n - 3) + terms$Taa / n / (n - 1) / (n - 2) / (n - 3)
      dsd <- sqrt(abs(dvar))
    } else {
      dvar <- terms$aijaij / n / n - 2 * terms$Saa / n / n / n  + terms$Taa / n / n / n / n
      dsd <-  sqrt(abs(dvar))
    }
    return(dsd) 

  }





#' Calculates the distance correlation \insertCite{szekely2007,szekely2009brownian}{dcortools}.
#'
#' @param X contains either the first  sample or its corresponding distance matrix.
#'
#' In the first case, X can be provided either as a vector (if one-dimensional), a matrix or a data.frame (if two-dimensional or higher). 
#'
#' In the second case, the input must be a distance matrix corresponding to the sample of interest.
#'
#' If X is a sample, type.X must be specified as "sample". If X is a distance matrix, type.X must be specified as "distance".
#' @param Y see X.
#' @param affine logical; specifies if the affinely invariant distance correlation \insertCite{dueck2014affinely}{dcortools} should be calculated or not.
#' @param standardize logical; specifies if X and Y should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.
#' @param bias.corr logical; specifies if the bias corrected version of the sample distance correlation \insertCite{huo2016fast}{dcortools} should be calculated.
#' @param type.X For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.
#' @param type.Y see type.X.
#' @param metr.X specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").
#' 
#'  Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
#'  
#'  For "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric, parameter)", e.g. c("gaussian", 3) for a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
#'  
#'  See \insertCite{lyons2013distance,sejdinovic2013equivalence,bottcher2017detecting;textual}{dcortools} for details.
#' @param metr.Y see metr.X.
#' @param use specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.
#' @param algorithm specifies the algorithm used for calculating the distance correlation. 
#' 
#' "fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 
#' 
#' "memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 
#' 
#' "standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.
#' 
#' "auto" chooses the best algorithm for the specific setting using a rule of thumb.
#' 
#' @return numeric; the distance correlation between samples X and Y.
#' @references
#' \insertRef{bottcher2017detecting}{dcortools}
#' 
#' \insertRef{dueck2014affinely}{dcortools}
#' 
#' \insertRef{huo2016fast}{dcortools}
#' 
#' \insertRef{lyons2013distance}{dcortools}
#' 
#' \insertRef{sejdinovic2013equivalence}{dcortools}
#' 
#' \insertRef{szekely2007}{dcortools}
#' 
#' \insertRef{szekely2009brownian}{dcortools}
#' @export
#' @examples 
#' X <- rnorm(200)
#' Y <- rnorm(200)
#' Z <- X + rnorm(200)
#' dim(X) <- dim(Y) <- dim(Z) <- c(20, 10)
#' 
#' #Demonstration that biased-corrected distance correlation is
#' #often more meaningful than without using bias-correction
#' 
#' distcor(X, Y) 
#' distcor(X, Z) 
#' distcor(X, Y, bias.corr = TRUE)
#' distcor(X, Z, bias.corr = TRUE)
#' 
#' #For more examples of the different options,
#' #see the documentation of distcov.
distcor <-
  function(X,
           Y,
           affine = FALSE,
           standardize=FALSE,
           bias.corr = FALSE,
           type.X = "sample",
           type.Y = "sample",
           metr.X = "euclidean",
           metr.Y = "euclidean",
           use = "all",
           algorithm = "auto") {
    #extract dimensions and sample sizes
    ss.dimX <- extract_np(X, type.X)
    ss.dimY <- extract_np(Y, type.Y)
    
    n <- ss.dimX$Sample.Size
    p <- ss.dimX$Dimension
    
    m <- ss.dimY$Sample.Size
    q <- ss.dimY$Dimension
    
    if  (use == "complete.obs") {
      ccX <- ccY <- cc <- 1:n
      ccX <- which(stats::complete.cases(X))
      ccY <- which(stats::complete.cases(Y))
      cc <- intersect(ccX, ccY)
      n <- m <- length(cc)
      if (type.X == "sample" && p == 1) {
        X <- X[cc]
      } else if (type.X == "sample" && p > 1) {
        X <- X[cc, ]
      }
      if (type.Y == "sample" && q == 1) {
        Y <- Y[cc]
      } else if (type.X == "sample" && q > 1) {
        Y <- Y[cc, ]
      }
      if (type.X == "distance") {
        X <- X[cc,cc]
      }
      if (type.Y == "distance") {
        Y <- Y[cc,cc]
      }
    }
    
    
    if (n != m) {
      stop("Samples X and Y must have the same sizes!")
    }
    
    
    ## normalize samples if calculation of affinely invariant distance correlation is desired
    if (affine == TRUE) {
      if (type.X == "distance" | type.Y == "distance") {
        stop("Affinely invariant distance correlation cannot be calculated for type distance")
      }
      if (p > n | q > n) {
        stop("Affinely invariant distance correlation cannot be calculated for p>n")
      }
      if (p > 1) {
        X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
      } else {
        X <- X / stats::sd(X)
      }
      if (q > 1) {
        Y <- Y %*% Rfast::spdinv(mroot(stats::var(Y)))
      } else {
        Y <- Y / stats::sd(Y)
      }
    } else if (standardize) {
      if (type.X == "distance" | type.Y == "distance") {
        stop("Standardization cannot be applied for type distance.")
      }
      if (p > 1) {
        X <- standardise(X, center = FALSE)
      } else {
        X <- X / stats::sd(X)
      }
      if (q > 1) {
        Y <- standardise(Y, center = FALSE)
      } else {
        Y <- Y / stats::sd(Y)
      }
    }
    
    
    alg.fast <- alg.standard <- alg.memsave <- FALSE
    
    if (algorithm == "auto") {
      if (p == 1 & q == 1 & metr.X[1] %in% c("euclidean", "discrete") 
          & metr.Y[1] %in% c("euclidean", "discrete") & type.X == "sample" & type.Y == "sample" & n > 100) {
        alg.fast <- TRUE
      } else if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") &
                 metr.Y[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") & type.X == "sample" & type.Y == "sample") {
        alg.memsave <- TRUE
      } else {
        alg.standard <- TRUE
      }
    } else if (algorithm == "fast") {
      alg.fast <- TRUE 
    } else if (algorithm == "standard") {
      alg.standard <- TRUE
    }  else if (algorithm == "memsave") {
      alg.memsave <- TRUE
    } else
      stop ("Algorithm must be one of \"fast\", \"standard\", \"memsave\" or \"auto\"")
    
    
    terms <- dcovterms(X, Y, n, calc.dcor = TRUE, doperm = FALSE, dobb3 = FALSE, alg.fast = alg.fast, alg.memsave = alg.memsave, alg.standard = alg.standard, p = p, q = q, metr.X = metr.X, metr.Y =metr.Y, type.X = type.X, type.Y = type.Y)
    
    
    if (bias.corr) {
      dvarX <- terms$aijaij / n / (n - 3) - 2 * terms$Saa / n / (n - 2) / (n - 3) + (terms$adotdot * terms$adotdot) / n / (n - 1) / (n - 2) / (n - 3)
      dvarY <- terms$bijbij / n / (n - 3) - 2 * terms$Sbb / n / (n - 2) / (n - 3) + (terms$bdotdot * terms$bdotdot) / n / (n - 1) / (n - 2) / (n - 3)
      dcov2 <- terms$aijbij / n / (n - 3) - 2 * terms$Sab / n / (n - 2) / (n - 3) + (terms$adotdot * terms$bdotdot) / n / (n - 1) / (n - 2) / (n - 3)
      dcov <-  sqrt(abs(dcov2)) * sign(dcov2)
      dcor <-  dcov / sqrt(sqrt(dvarX * dvarY))
    } else {
      dvarX <- terms$aijaij / n / n - 2 * terms$Saa / n / n / n + (terms$adotdot * terms$adotdot) / n / n / n / n
      dvarY <- terms$bijbij / n / n - 2 * terms$Sbb / n / n / n + (terms$bdotdot * terms$bdotdot) / n / n / n / n
      dcov2 <- terms$aijbij / n / n - 2 * terms$Sab / n / n / n  + (terms$adotdot * terms$bdotdot) / n / n / n / n
      dcov <- sqrt(abs(dcov2))
      dcor <- dcov / sqrt(sqrt(dvarX * dvarY))
    }
    
 
    return(dcor)
  }

centmat <- function(X,
                    metr.X = "euclidean",
                    p) {
  
  distX <- distmat(X, metr.X, p)
  return(doublecent(distX))
}



doublecent <- function(distX) {
  n <- nrow(distX)
  rm <- Rfast::rowmeans(distX)
  totm <- mean(rm)
  rmmat <- matrix(rep(rm,n), ncol=n)
  centmat <- distX - rmmat - t(rmmat) + totm
  return(centmat)
}




distmat <- function(X,
                    metr.X = "euclidean",
                    p) {
  len <- length(metr.X) 
  
  if (len == 1) {
    if (p == 1) {
      X <- as.numeric(X)
      if (metr.X == "euclidean" | metr.X == "alpha" | metr.X == "minkowski") {
        distX <- Rfast::vecdist(X)
      } else if (metr.X == "gaussian") {
        distX <- 1 - exp(-0.5 * Rfast::vecdist(X) ^ 2)
      } else if (metr.X == "gaussauto") {
        preX <- Rfast::vecdist(X)^2
        bw <- stats::median(stats::as.dist(preX), na.rm = TRUE)
        distX <- 1 - exp(- preX / bw)
      } else if (metr.X == "boundsq") {
        distX <- Rfast::vecdist(X) 
        distX <- (distX ^ 2) / (1 + distX ^ 2)
      } else if (metr.X == "discrete") {
        distX <- 1 * (Rfast::vecdist(X)>0)
      } else {
        distfunc <- match.fun(metr.X)
        distX <- distfunc(X = X)
        if (formals(distfunc)$kernel) {
          distX <- kerntodist(distX)
        }
      } 
    } else {
      if (metr.X == "euclidean" | metr.X == "alpha" | metr.X == "minkowski") {
        distX <- Rfast::Dist(X)
      } else if (metr.X == "gaussian") {
        distX <- 1 - exp(-0.5 * Rfast::Dist(X) ^ 2)
      } else if (metr.X == "gaussauto") {
        preX <- Rfast::Dist(X)^2
        bw <- stats::median(stats::as.dist(preX), na.rm= TRUE)
        distX <- 1 - exp(- preX / bw)
      } else if (metr.X == "boundsq") {
        distX <- Rfast::Dist(X) 
        distX <- (distX ^ 2) / (1 + distX ^ 2)
      } else if (metr.X == "discrete") {
        distX <- 1*(Rfast::Dist(X)>0)
      } else {
        distfunc <- match.fun(metr.X)
        distX <- distfunc(X = X)
        if (formals(distfunc)$kernel) {
          distX <- kerntodist(distX)
        }
      } 
    }
  } else {
    prm <- as.numeric(metr.X[2:len])
    metr.X <- metr.X[1]
    if (p == 1) {
      if (metr.X == "alpha") {
        distX <- Rfast::vecdist(X) ^ prm
      } else if (metr.X == "gaussian") {
        distX <- 1 - exp(-0.5 * Rfast::vecdist(X) ^ 2 / prm ^ 2)
      } else if (metr.X == "gaussauto") {
        preX <-Rfast::vecdist(X)^2
        bw <- stats::median(stats::as.dist(preX), na.rm= TRUE)
        distX <- 1 - exp(- preX / bw / prm ^ 2)
        } else if (metr.X == "boundsq") {
        distX <- Rfast::vecdist(X) 
        distX <- (distX ^ 2) / (prm ^ 2 + distX ^ 2)
      } else if (metr.X == "minkowski") {
        distX <- Rfast::vecdist(X) 
      } else {
        distfunc <- match.fun(metr.X)
        distX <- distfunc(X = X, prm = prm)
        if (formals(distfunc)$kernel) {
          distX <- kerntodist(distX)
        }
      } 
    } else {
      if (metr.X == "alpha") {
        distX <- Rfast::Dist(X) ^ prm
      } else if (metr.X == "gaussian") {
        distX <- 1 - exp(-0.5 * Rfast::Dist(X) ^ 2 / prm ^ 2)
      } else if (metr.X == "gaussauto") {
        preX <-Rfast::Dist(X)^2
        bw <- stats::median(stats::as.dist(preX), na.rm= TRUE)
        distX <- 1 - exp(- preX / bw / prm ^ 2)
      } else if (metr.X == "boundsq") {
        distX <- Rfast::Dist(X) 
        distX <- (distX ^ 2) / (prm ^ 2 + distX ^ 2)
      } else if (metr.X == "minkowski") {
        if (prm > 2 | prm <= 1)
          warning("Distance covariance with Minkowski distance outside the interval (1,2] does not define independence")
        distX <- as.matrix(stats::dist(X, method="minkowski", p=prm)) 
      } else {
        distfunc <- match.fun(metr.X)
        distX <- distfunc(X = X, prm = prm)
        if (formals(distfunc)$kernel) {
          distX <- kerntodist(distX)
        }
      } 
    }
  }
  return(distX)
}


kerntodist <- function(kernmat) {
  n <- ncol(kernmat)
  dmat <- matrix(rep(diag(kernmat), n), ncol = n)
  return((dmat + t(dmat) - 2 * kernmat) / 2)
}


#' Extract the dimensions of X.
#'
#' @param X a numeric vector or a numeric matrix.
#' @param type.X either "sample" or "distance". If type.X = "sample", X must be
#' a numeric vector or numeric matrix with the corresponding observations. If metr.X = "distance",
#' X must be a distance matrix.
#'
#' @return The centralized distance matrix corresponding to X.
extract_np <- function(X, type.X) {
  if (type.X == "sample") {
    if (is.vector(X) | is.factor(X))
    {
      n <- length(X)
      p <- 1L
    } else if (is.matrix(X) | is.data.frame(X)) {
      n <- nrow(X)
      p <- ncol(X)
    } else {
      stop("X must be a vector, matrix or dataframe for type 'sample'!")
    }
  } else if (type.X == "distance") {
    if (is.matrix(X)) {
      n <- nrow(X)
      p <- NA
    } else {
      stop("X must be a matrix for type 'distance'!")
    }
  } else {
    stop("type.X must be either 'sample' or 'distance'.")
  }
  return(list("Sample.Size" = n, "Dimension" = p))
}


decom.metr<- function(metr.X) {
  if (length(metr.X) == 1) {
    prmX <- 0
  } else {
    prmX <- as.numeric(metr.X[2])
    metr.X <- as.character(metr.X[1])
    if (is.na(prmX)) 
      stop("Parameters of standard metrics must be numeric")
  }
  return(c(metr.X,prmX))
}

normalize.sample <- function(X, n, p) {
  if (p > n) {
    stop("Affinely invariant distance covariance cannot be calculated for p>n")
  }
  if (p > 1) {
    X <- X %*% Rfast::spdinv(mroot(stats::var(X)))
  } else {
    X <- X / stats::sd(X)
  }
}

#' @noRd
scale.sample <- function(X, n, p) {
  if (p > 1) {
    X <- standardise(X, center = FALSE)
  } else {
    X <- X / stats::sd(X)
  }
}

mroot <- function(A) {
  e <- eigen(A)
  V <- e$vectors
  V %*% diag(e$values) %*% t(V)
  
  
  B <- V %*% diag(sqrt(e$values)) %*% t(V)
  return(B)
}

dcovterms <- function(X,Y,n, calc.dcor, doperm, dobb3, alg.fast, alg.memsave, alg.standard, p, q, metr.X, metr.Y, type.X, type.Y) {

    if (alg.fast) {
    if (p == 1 & q == 1) {
      if (metr.X == "euclidean" & metr.Y == "euclidean") {
        terms <- dcovterms.fast(X, Y, n, calc.dcor = calc.dcor, calc.perm = doperm)
      } else if (metr.X == "euclidean" & metr.Y == "discrete") {
        Y <- as.factor(Y)
        terms <- dcovterms.fast.numdisc(X, Y, n, calc.dcor = calc.dcor, calc.perm = doperm)
      } else if (metr.X == "discrete" & metr.Y == "euclidean") {
        X <- as.factor(X)
        terms <- dcovterms.fast.numdisc(Y, X, n, calc.dcor = calc.dcor, calc.perm = doperm)
      } else if (metr.X == "discrete" & metr.Y == "discrete") {
        X <- as.factor(X)
        Y <- as.factor(Y)
        terms <- dcovterms.fast.discrete(X ,Y, n, calc.dcor = calc.dcor, calc.perm = doperm)
      } else {
        stop("metr.X and metr.Y have to be \"euclidean\" or \"discrete\" for fast algorithms")
      }
    } else {
      stop("Dimensions of X and Y must be 1 for fast algorithms.")
    }
  } else if (alg.memsave) {
    if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") &
        metr.Y[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete")) {
      terms <- dcovterms.memsave(X, Y, metr.X, metr.Y, p, q, calc.dcor = calc.dcor, calc.perm = doperm)
    } else {
      stop("Memory efficient algorithms cannot be run with user-defined metrics")
    }
  } else if (alg.standard) {
    terms <- dcovterms.standard(X, Y, type.X, type.Y, metr.X, metr.Y, p, q, calc.dcor = calc.dcor, calc.bb3 = dobb3 | doperm)
}
 return(terms)  
}






dcovterms.standard <- function(X, Y = NULL, type.X, type.Y = NULL, metr.X, metr.Y = NULL, p, q = NULL, calc.dvar = FALSE, calc.dcor = FALSE, calc.bb3 = FALSE) {
  ## if distance matrix is given
  if (type.X == "distance") {
    distX <- X
  } else {
    distX <- distmat(X, metr.X, p)
  }
  
  aidot <- Rfast::colsums(distX)
  adotdot <- sum(aidot)
  
  if (calc.dvar) {
    aijaij <- matrix_prod_sum(distX, distX)
    Saa <- vector_prod_sum(aidot, aidot)
    Taa <- adotdot * adotdot
    return(list("aijaij" = aijaij, "Saa" = Saa, "Taa" = Taa))
  }
  
  if (type.Y == "distance") {
    distY <- Y
  } else {
    distY <- distmat(Y, metr.Y, q)
  }
  

  bidot <- Rfast::colsums(distY)
  bdotdot <- sum(bidot)
  
  aijbij <- matrix_prod_sum(distX, distY)
  Sab <- vector_prod_sum(aidot, bidot)
  
  if (calc.dcor) {
    aijaij <- matrix_prod_sum(distX, distX)
    bijbij <- matrix_prod_sum(distY, distY)
    Saa <- vector_prod_sum(aidot, aidot)
    Sbb <- vector_prod_sum(bidot, bidot)
    if (calc.bb3)
      return(list("aijbij" = aijbij, "Sab" = Sab, "aijaij" = aijaij, "Saa" = Saa, "adotdot" = adotdot, "bijbij" = bijbij, "Sbb" = Sbb, "bdotdot" = bdotdot, "aidot" = aidot, "bidot" = bidot, "distX" = distX, "distY" = distY))
      else
      return(list("aijbij" = aijbij, "Sab" = Sab, "aijaij" = aijaij, "Saa" = Saa, "adotdot" = adotdot, "bijbij" = bijbij, "Sbb" = Sbb, "bdotdot" = bdotdot))
  }
  

  
  Tab <- adotdot * bdotdot
  
  return(list("aijbij" = aijbij, "Sab" = Sab, "Tab" = Tab))
}



dcovterms.fast <- function(X, Y = NULL, n, calc.dvar = FALSE, calc.dcor = FALSE, calc.perm = FALSE) {
  X <- as.numeric(X)
  Y <- as.numeric(Y)
  
  temp <- IX <- IY <- IY.s <- 1:n
  #
  
  IX0 <- Rfast::Order(X)
  X.sort.X <- X[IX0]
  IX[IX0] <- temp
  
  sX.X <- cumsum(X.sort.X)
  alphaX <-  0:(n - 1)
  betaX <- sX.X - X.sort.X
  Xdot <- sX.X[n]
  aidot <- Xdot + (2 * alphaX - n) * X.sort.X - 2 * betaX
  adotdot <- sum(aidot)
  
  if (calc.dvar) {
    XXdot <- sum(X ^ 2)
    aijaij <- 2 * n * XXdot - 2* Xdot^2
    Saa <- vector_prod_sum(aidot, aidot)
    Taa <- adotdot * adotdot
    return(list("aijaij" = aijaij, "Saa" = Saa, "Taa" = Taa))
  }
  
  IY0 <- Rfast::Order(Y)
  Y.sort.Y <- Y[IY0]
  IY[IY0] <- temp
  alphaY <- 0:(n - 1)
  
  sY.Y <- cumsum(Y.sort.Y)
  
  betaY <- sY.Y - Y.sort.Y
  
  Ydot <- sY.Y[n]
  
  bidot <- Ydot + (2 * alphaY - n) * Y.sort.Y - 2 * betaY
  bdotdot <- sum(bidot)
  
  Y.sortX <- Y[IX0]
  XY.sortX <- Y.sortX * X.sort.X
  IY0.s <- IX[IY0]
  IY.s[IY0.s] <- temp
  
  
  sY.X <- cumsum(X.sort.X[IY0.s])
  sX.Y <- cumsum(Y.sortX)
  sY.XY <- cumsum(XY.sortX[IY0.s]) 
  sX.XY <- cumsum(XY.sortX)
  
 
  aijbij <- SUMAIJBIJ(IY.s, X.sort.X, Y.sortX, XY.sortX, sX.X, sX.Y, sX.XY, sY.X[IY.s], sY.Y[IY.s], sY.XY[IY.s])
 
  Sab <- vector_prod_sum(aidot, bidot[IY.s])
  
  if (calc.dcor) {
    XXdot <- sum(X^2)
    YYdot <- sum(Y^2)
    Saa <- vector_prod_sum(aidot, aidot)
    aijaij <- 2 * n * XXdot - 2* Xdot^2
    Sbb <- vector_prod_sum(bidot, bidot)
    bijbij <- 2 * n * YYdot - 2* Ydot^2
    if (calc.perm) {
      aidot <- aidot[IX]
      bidot <- bidot[IY]
      return(list("aijbij" = aijbij, "Sab" = Sab, "aijaij" = aijaij, "Saa" = Saa, "adotdot" = adotdot, "bijbij" = bijbij, "Sbb" = Sbb, "bdotdot" = bdotdot, 
                  "aidot" = aidot, "bidot" = bidot, "X.sort.X" = X.sort.X, "Y" = Y, "IX0" = IX0, "sX.X" = sX.X, "sY.Y" = sY.Y, "IY" = IY, "n" = n, "ndisc"=0))
    }
      else
      return(list("aijbij" = aijbij, "Sab" = Sab, "aijaij" = aijaij, "Saa" = Saa, "adotdot" = adotdot, "bijbij" = bijbij, "Sbb" = Sbb, "bdotdot" = bdotdot))
  }
  
  
  Tab <- adotdot * bdotdot
  
  return(list("aijbij" = aijbij, "Sab" = Sab, "Tab" = Tab))
}


dcovterms.fast.discrete <- function(X, Y = NULL, n, calc.dvar = FALSE, calc.dcor = FALSE, calc.perm = FALSE) {
  
  if (calc.dvar) {
    labX <- levels(X)
    fX <- length(labX)
    nX <- table(X)
    aijaij <- n * (n - 1) - sum(nX * (nX - 1)) 
    aidot <- as.numeric(n - nX)
    adotdot <- sum(aidot * nX)
    Saa <- sum(aidot * aidot * nX)
    Taa <- adotdot*adotdot
    return(list("aijaij" = aijaij, "Saa" = Saa, "Taa" = Taa))
  } 
  
  nXY <- Rfast::Table(X, Y, names = FALSE)
  nX <- Rfast::rowsums(nXY)
  nY <- Rfast::colsums(nXY)
  
  
  aidot <- as.numeric(n - nX)
  adotdot <- sum(aidot * nX)
  bidot <- as.numeric(n - nY)
  bdotdot <- sum(bidot * nY)
  
  aijbij <- n * n - sum(nX * nX) - sum(nY * nY) + sum(nXY * nXY)
  Sab <- sum((aidot %*% t(bidot)) * nXY)
  
  
  if (calc.dcor) {
    aijaij <- n * (n - 1) - sum(nX * (nX - 1)) 
    bijbij <- n * (n - 1) - sum(nY * (nY - 1)) 
    Saa <- sum(aidot * aidot * nX)
    Sbb <- sum(bidot * bidot * nY)
    if (calc.perm) {
      return(list("aijbij" = aijbij, "Sab" = Sab, "aijaij" = aijaij, "Saa" = Saa, "adotdot" = adotdot, "bijbij" = bijbij, "Sbb" = Sbb, "bdotdot" = bdotdot,
                  "X" = X, "Y" = Y, "nX" = nX, "nY" = nY, "aidot" = aidot, "bidot" = bidot, "n" = n,"ndisc"=2)) 
    } else {
      return(list("aijbij" = aijbij, "Sab" = Sab, "aijaij" = aijaij, "Saa" = Saa, "adotdot" = adotdot, "bijbij" = bijbij, "Sbb" = Sbb, "bdotdot" = bdotdot))
    }
  } 
  
  Tab <- adotdot*bdotdot
  
  return(list("aijbij" = aijbij, "Sab" = Sab, "Tab" = Tab))
  
}


dcovterms.fast.numdisc <- function(X, Y, n, calc.dcor =FALSE, calc.perm = FALSE) {
  IX0 <- Rfast::Order(X)
  X.sort.X <- X[IX0]

  
  sX.X <- cumsum(X.sort.X)
  alphaX <- 0:(n - 1)
  betaX <- sX.X - X.sort.X
  Xdot <- sX.X[n]
  Y.sortX <- Y[IX0]
  
  aidot <- Xdot + (2 * alphaX - n) * X.sort.X - 2 * betaX
  adotdot <- sum(aidot)
  levY <- levels(Y)
  nY <- Rfast::Table(Y, names = FALSE)
  aijbij <- adotdot
  
  
  for (lvl in levY) {
    set0 <- which(Y.sortX == lvl)
    n0 <- length(set0)
    alphaX0 <- 0:(n0 - 1)
    X0 <- X.sort.X[set0]
    sX.X0 <- cumsum(X0)
    betaX0 <- sX.X0 - X0
    X0dot <- sX.X0[n0]
    aidot0 <- X0dot + (2 * alphaX0 - n0) * X0 - 2 * betaX0
    adotdot0 <- sum(aidot0)
    aijbij <- aijbij - adotdot0
  }
  
  bidot <- as.numeric(n - nY[Y.sortX])
  bdotdot <- sum(bidot)
  
  Sab <- vector_prod_sum(aidot, bidot)
  
  
  if (calc.dcor) {
    XXdot <- sum(X ^ 2)
    aijaij <- 2 * n * XXdot - 2 * Xdot ^ 2
    bijbij <- n * (n - 1) - sum(nY * (nY - 1))
    Saa <- sum(aidot * aidot)
    Sbb <- sum(bidot * bidot)
    if (calc.perm)
      return(
        list(
          "aijbij" = aijbij,
          "Sab" = Sab,
          "aijaij" = aijaij,
          "Saa" = Saa,
          "adotdot" = adotdot,
          "bijbij" = bijbij,
          "Sbb" = Sbb,
          "bdotdot" = bdotdot,
          "aidot" = aidot,
          "bidot" = bidot,
          "X.sort.X" = X.sort.X,
          "Y.sortX" = Y.sortX,
          "levY" = levY,
          "n" = n,
          "ndisc"=1
        )
      )
    else
      return(
        list(
          "aijbij" = aijbij,
          "Sab" = Sab,
          "aijaij" = aijaij,
          "Saa" = Saa,
          "adotdot" = adotdot,
          "bijbij" = bijbij,
          "Sbb" = Sbb,
          "bdotdot" = bdotdot
        )
      )
  } 
  
  
  Tab <- adotdot * bdotdot
  
  return(list("aijbij" = aijbij, "Sab" = Sab, "Tab" = Tab))
}


dcovterms.memsave <-  function(X, Y, metr.X, metr.Y, p, q, calc.dcor = FALSE, calc.perm = FALSE) {
  
  if (length(metr.X) == 1) {
    prmX <- 0
  } else {
    prmX <- as.numeric(metr.X[2])
    metr.X <- as.character(metr.X[1])
    if (is.na(prmX)) 
      stop("Parameters of standard metrics must be numeric")
  }
  
  if (length(metr.Y) == 1) {
    prmY <- 0
  } else {
    prmY <- as.numeric(metr.Y[2]) 
    metr.Y <- metr.Y[1]
    if (is.na(prmY)) 
      stop("Parameters of standard metrics must be numeric")
  }
  
  
  
  if (p == 1 & q == 1) {
    if (metr.X == "discrete") {
      X <- as.numeric(X)
    }
    if (metr.Y == "discrete") {
      Y <- as.numeric(Y)
    }
    terms <- dcovtermsmemvec(X, Y, metr.X, metr.Y, prmX, prmY, calcdcor = calc.dcor, calcperm = calc.perm)
  } else if (p > 1 & q > 1) {
    terms <- dcovtermsmem(X, Y, metr.X, metr.Y, prmX, prmY, calcdcor = calc.dcor, calcperm = calc.perm)
  } else if (p > 1) {
    if (metr.Y == "discrete") {
      Y <- as.numeric(Y)
    }
    Y <- as.matrix(Y)
    terms <- dcovtermsmem(X, as.matrix(Y), metr.X, metr.Y, prmX, prmY, calcdcor = calc.dcor, calcperm = calc.perm)
  } else {
    if (metr.X == "discrete") {
      X <- as.numeric(X)
    }
    X <- as.matrix(X)
    terms <- dcovtermsmem(X, Y, metr.X, metr.Y, prmX, prmY, calcdcor = calc.dcor, calcperm = calc.perm)
  }
  
  if (calc.perm) {
    terms$prmX <- prmX
    terms$prmY <- prmY
    terms$metr.X <- metr.X
    terms$metr.Y <- metr.Y
    terms$p <- p
    terms$q <- q
    terms$X <- X
    terms$Y <- Y
  }
  
  return(terms)
}  



dvarterms.memsave <-  function(X, metr.X, p) {
  
  if (length(metr.X) == 1) {
    prmX <- 0
  } else {
    prmX <- as.numeric(metr.X[2])
    metr.X <- as.character(metr.X[1])
    if (is.na(prmX)) 
      stop("Parameters of standard metrics must be numeric")
  }
  

  if (p == 1) {
    if (metr.X == "discrete") {
      X <- as.numeric(X)
    }
    terms <- dvartermsmemvec(X, metr.X, prmX)
  } else {
    terms <- dvartermsmem(X, metr.X, prmX)
  }
  
  terms$Saa <- vector_prod_sum(terms$aidot,terms$aidot)
  terms$Taa <- terms$adotdot * terms$adotdot
  
  return(terms)
}  


aijbij.memsave <-  function(X, Y, metr.X, metr.Y, p, q) {
  
  if (length(metr.X) == 1) {
    prmX <- 0
  } else {
    prmX <- as.numeric(metr.X[2])
    metr.X <- as.character(metr.X[1])
    if (is.na(prmX)) 
      stop("Parameters of standard metrics must be numeric")
  }
  
  if (length(metr.Y) == 1) {
    prmY <- 0
  } else {
    prmY <- as.numeric(metr.Y[2]) 
    metr.Y <- metr.Y[1]
    if (is.na(prmY)) 
      stop("Parameters of standard metrics must be numeric")
  }
  

  if (p == 1 & q == 1) {
    if (metr.X == "discrete") {
      X <- as.numeric(X)
    }
    if (metr.Y == "discrete") {
      Y <- as.numeric(Y)
    }
    terms <- aijbijmemvec(X, Y, metr.X, metr.Y, prmX, prmY)
  } else if (p > 1 & q > 1) {
    terms <- aijbijmem(X, Y, metr.X, metr.Y, prmX, prmY)
  } else if (p > 1) {
    if (metr.Y == "discrete") {
      Y <- as.numeric(Y)
    }
    Y <- as.matrix(Y)
    terms <- aijbijmem(X, Y, metr.X, metr.Y, prmX, prmY)
  } else {
    if (metr.X == "discrete") {
      X <- as.numeric(X)
    }
    X <- as.matrix(X)
    terms <- aijbijmem(X, Y, metr.X, metr.Y, prmX, prmY)
  }
  
  
  return(terms)
}  

#' Calculates distance correlation from Pearson correlation under assumption of a bivariate normal distribution
#' @param rho Pearson correlation.
#' @return Distance correlation assuming a bivariate normal distribution
#' @export
dcorgaussianbiv <- function(rho)
{
  nenner <-   rho * asin(rho) + sqrt(1-rho^2) - rho * asin(rho/2) - sqrt(4-rho^2) +1
  zaehler <- 1+ pi/3 - sqrt(3)
  return(sqrt(nenner/zaehler))
}


#' @export
print.dcmatrix <- function(x, ...) {
  dcmat <- x
  if (dcmat$test == "none")
    testout <- "None"
    else if (dcmat$test == "permutation")
    testout <- paste("Permutation test with ",dcmat$b," permutations")
    else if (dcmat$test == "gamma")
    testout <- "Simple gamma approximation"               
    else if (dcmat$test == "conservative")
    testout <- "Conservative two-moment approximation"  
    else if (dcmat$test == "bb3")
    testout <- "Three moment approximation by Berschneider and Boettcher"
    
    
    if (dcmat$test.pw == "none")
      testout.pw <- "None"
    else if (dcmat$test.pw == "permutation")
      testout.pw  <- paste("Permutation test with ",dcmat$b," permutations")
    else if (dcmat$test.pw == "gamma")
      testout.pw  <- "Simple gamma approximation"               
    else if (dcmat$test.pw == "conservative")
      testout.pw  <- "Conservative two-moment approximation"  
    else if (dcmat$test.pw == "bb3")
      testout.pw  <- "Three moment approximation by Berschneider and Boettcher"
    
    pw <- (dcmat$test.pw != "none") | dcmat$calc.dcor.pw | dcmat$calc.dcov.pw
    
    if (dcmat$calc.cor == "pearson")
      corout <- "Pearson"
    else if (dcmat$calc.cor == "spearman")
    corout <- "Spearman"
    else if (dcmat$calc.cor == "kendall")
    corout <- "Kendall"               
    else 
    corout <- "None"     
    
  dX <- dY <- dcmat$dX
  dX.pw <- dY.pw <- dcmat$dX.pw
    
  if (dcmat$withY) {  
    dY.pw <- dcmat$dY.pw
    dY <- dcmat$dY
  }
  cat(paste("dcmatrix object of size", dX, "x", dY, "\n"))
  cat("\n")
  cat("Call:", "\n")
  cat(paste(deparse(dcmat$call), "\n"))
  cat("\n")
  cat(paste("Distance correlation matrix: ", ifelse(dcmat$calc.dcor,"Available \n", "Not available \n")) )
  cat(paste("Distance covariance matrix: ", ifelse(dcmat$calc.dcov,"Available \n", "Not available \n")) )
  cat(paste("Bias correction: ", ifelse(dcmat$bias.corr,"Yes \n", "No \n")) )
  cat(paste("Affinely invariant: ", ifelse(dcmat$affine,"Yes \n", "No \n")) )
  cat(paste("Test: ", testout, "\n") )
  cat(paste("Standard correlation matrix: ", corout, "\n"))
  if (pw) {
    cat("\n")
    cat("Pairwise results:\n")
    cat("\n")
    cat(paste("Size:", dX.pw, "x", dY.pw, "\n"))
    cat("\n")
    cat(paste("Distance correlation matrix: ", ifelse(dcmat$calc.dcor.pw,"Available \n", "Not available \n")) )
    cat(paste("Distance covariance matrix: ", ifelse(dcmat$calc.dcov.pw,"Available \n", "Not available \n")) )
    cat(paste("Test: ", testout.pw, "\n") )
  }
}

#' Plots a heatmap from a dcmatrix object using the function "pheatmap" from the package "pheatmap".
#'
#' @param x a dcmatrix object.
#' @param type specifies what should be displayed in the heatmap. One of "dcor", "dcov", "logp" (-log10 of corresponding p-values), "cor", "abscor" (absolute correlation), "logp.cor", "dcor.pw", "dcov.pw" or "logp.pw".
#' @param trunc.up truncates the values to be plotted; if set to numeric, all values larger than trunc.up are set to trunc.up.
#' @param trunc.low truncates the values to be plotted; if set to numeric, all values smaller than trunc.low are set to trunc.low.
#' @param cluster_rows,cluster_cols,display_numbers passed to pheatmap().
#' @param ... passed to pheatmap().
#' @return a heatmap plotting the entries of the slot specified in type of the object specified in dcmat.
#' @export
plot.dcmatrix <- function(x, type = "dcor", trunc.up = NULL, trunc.low = NULL, cluster_rows = FALSE, cluster_cols = FALSE, display_numbers=TRUE,  ...) {
  dcmat <- x
  if (type =="dcor") {
    mat <- dcmat$dcor 
  } else if (type == "dcov") {
    mat <- dcmat$dcov 
  } else if (type == "logp") {
    mat <- -log10(dcmat$pvalue)
  } else if (type == "cor") {
    mat <- dcmat$cor
  } else if (type == "abscor") {
    mat <- abs(dcmat$cor)
  } else if (type == "logp.cor") {
    mat <- -log10(dcmat$pval.cor)
  } else if (type == "dcor.pw") {
    mat <- dcmat$dcor.pw 
  } else if (type == "dcov.pw") {
    mat <- dcmat$dcov.pw 
  } else if (type == "logp.pw") {
    mat <- -log10(dcmat$pvalue.pw)
  } else {
    stop(" \"type\" must be one of \"dcor\",\"logp\",\"cor\",\"abscor\", \"logp.cor\",\"dcor.pw\",\"dcov.pw\" or \"logp.pw\" ")
  }
  
  if (!is.null(trunc.up) | !is.null(trunc.low)) {
    mat[mat > trunc.up] <- trunc.up
    mat[mat < trunc.low] <- trunc.low
  }  
  
  if (anyNA(mat)|any(mat==Inf)) {
    warning("NAs and Infs were set to 0 for plotting.")
    mat[is.na(mat)] <- 0
    mat[mat == Inf] <- 0
  }
  

  pheatmap::pheatmap(mat, cluster_rows = cluster_rows, cluster_cols = cluster_cols, display_numbers= display_numbers, ...)

}



#' Plots Pearson/Spearman/Kendall correlation against distance correlation (often resembling a horseshoe(hs)).
#'
#' @param dcmat A dcmatrix object.
#' @param maxcomp Maximum number of associations, for which distance correlation is plotted against correlation. If the number of associations in the dcmat object is larger, only the maxcomp associations with the largest difference between distance correlation and absolute (Pearson/Spearman/Kendall) correlation are plotted.
#' @param col color of the plot.
#' @param alpha alpha parameter of the plot.
#' @param cortrafo Either "none" or "gaussiandcor". If "gaussiandcor", the distance correlation under assumption of normality is calculated and plotted against the actual distance correlation. 
#' 
#' Note that this is only sensible for Pearson correlation!
#' @return Plot of (possibly transformed) Pearson/Spearman/Kendall correlation against distance correlation.
#' @export
hsplot <- function(dcmat, maxcomp = 1e5, col = "blue", alpha = 1, cortrafo = "none") {
  pw <- FALSE
  
  if (is.null(dcmat$corr) | is.null(dcmat$dcor))
    stop("Correlation and (pairwise) distance correlation matrix needed for horseshoe plot!")

  if (anyDuplicated(dcmat$group.X)) {
   pw <- TRUE
   if (is.null(dcmat$dcor.pw))
     stop("Pairwise distance correlation matrix needed for horseshoe plot!")
  }#
  
  if (dcmat$withY) {
    if (anyDuplicated(dcmat$group.Y)) {
      pw <- TRUE
      if (is.null(dcmat$dcor.pw))
        stop("Pairwise distance correlation matrix needed for horseshoe plot!")
    }
  }
    
  
  

  if (!dcmat$withY) {
    corr <- c(stats::as.dist(dcmat$corr))
    if (pw)
      dcor <- c(stats::as.dist(dcmat$dcor.pw))
    else
      dcor <- c(stats::as.dist(dcmat$dcor))
  } else {
    
    corr <- c(dcmat$corr)
    if (pw)
      dcor <- c(dcmat$dcor.pw)
    else 
      dcor <- c(dcmat$dcor)
  }
  
  
  
  xl <- "Correlation"

  if (cortrafo == "gaussiandcor") {
    corr <- dcorgaussianbiv(corr)
    xl <- "Distance Correlation under asssumption of bivariate normal"
  }

  if (length(corr) > maxcomp) {
    sel <- order(dcor-abs(corr), decreasing=TRUE)[1:maxcomp]
    warning("Not all comparisons plotted. Increase argument maxcomp to view all comparisons")
    dcor <- dcor[sel]
    corr <- corr[sel]
  }
  print(ggplot2::ggplot() + ggplot2::geom_point(ggplot2::aes(dcor,corr), alpha = alpha, color =col) + ggplot2::ylab(xl) + ggplot2::xlab("Distance Correlation"))
}

#' #' @export
#' plotrelations <- function(dcmat, candidates = NULL, diff, smooth = "loess", cortrafo = "gaussiandcor", switchXY = FALSE, names = "nonlin", width = 480, height = 480, criterion = "dcor") {
#'   if (is.null(dcmat$X)) {
#'     stop("Data needed to plot interesting nonlinear associations! Run dcmatrix again or add data manually to dcmat object.")
#'   } 
#'   
#'   X <- as.matrix(dcmat$X)
#'   
#'   if (dcmat$withY) 
#'     Y <- as.matrix(dcmat$Y)
#'   
#'   if (criterion == "logp") {
#'     if (is.null(dcmat$pvalue) | is.null(dcmat$pval.cor)) {
#'       stop("P-values missing in dcmatrix object.")
#'     } 
#'     corr <- -log10(dcmat$pval.cor)
#'     dcor <- -log10(dcmat$pvalue)
#'   } else if (criterion == "dcor") {
#'     if (is.null(dcmat$dcor) | is.null(dcmat$corr)) {
#'       stop("Distance correlation or correlation missing in dcmatrix object.")
#'     }
#'     corr <- abs(dcmat$corr)
#'     dcor <- dcmat$dcor
#'   }
#' 
#'   
#'   if (!dcmat$withY) {  
#'     lowtr <- lower.tri(corr)
#'     corr[lowtr] <- NA
#'     dcor[lowtr] <- NA
#'     Y <- X
#'   }
#'   
#'   if (cortrafo == "gaussiandcor") {
#'     corr <- dcorgaussianbiv(corr)
#'     xl <- "Distance Correlation under asssumption of bivariate normal"
#'   }  
#'   
#'   if (is.null(candidates)) {
#'     candidates <- which((dcor - corr) > diff, arr.ind=TRUE)
#'   }
#'   ncand <- nrow(candidates)
#'   
#'   
#'   if (ncand<0.5) {
#'     stop("No associations with specified difference between two tests/association measures found.")
#'   }
#'   
#'   names.Y <- "Y"
#'   
#'   if (is.null(colnames(X)))
#'     colnames(X) <- paste("X",1:ncol(X))
#'   
#'   
#'   
#'   
#'   
#'   if (!dcmat$withY) {  
#'     Y <- X
#'     names.Y <- "X"
#'     if (switchXY) {
#'       candidates <- cbind(candidates[,2],candidates[,1])
#'     }
#'   } else {
#'     if (dcmat$withY) {
#'       if (is.null(colnames(Y)))
#'         colnames(Y) <- paste("Y",1:ncol(Y), sep = "")
#'     }
#'     if (switchXY) {
#'       Xnew <- Y
#'       Ynew <- X
#'       X <- Xnew
#'       Y <- Ynew
#'       candidates <- cbind(candidates[,2],candidates[,1])
#'     }
#'   }
#'   
#'   
#'   for (i in 1:ncand) {
#'     name.X <- colnames(X)[candidates[i,1]]
#'     name.Y <- colnames(Y)[candidates[i,2]]
#'     
#'     title <- paste("Association of ", name.X," with ", name.Y,sep="")
#'     filename <- paste(names,"X-",name.X,"_Y-",name.Y,".png",sep="")
#'     
#'     png(filename, width = width, height = height)
#'     plotobj <- ggplot2::ggplot()+ggtitle(title)+ ggplot2::xlab(name.X) + ggplot2::ylab(name.Y)
#'     plotobj <- plotobj + ggplot2::geom_point(ggplot2::aes(X[,candidates[i,1]], Y[,candidates[i,2]]))
#'     if (smooth != "none")
#'       plotobj <- plotobj + geom_smooth(ggplot2::aes(X[,candidates[i,1]], Y[,candidates[i,2]]), method=smooth, se=F)  
#'     print(plotobj)
#'     dev.off()
#'   }
#' }
#' 
#' #' @export
#' extract.nonlins <- function(dcmat, diff, criterion = "dcor") {
#'   if (is.null(dcmat$X)) {
#'     stop("Data needed to plot interesting nonlinear associations! Run dcmatrix again or add data manually to dcmat object.")
#'   } 
#'   
#'   if (criterion == "logp") {
#'     if (is.null(dcmat$pvalue) | is.null(dcmat$pval.cor)) {
#'       stop("P-values missing in dcmatrix object.")
#'     } 
#'     corr <- -log10(dcmat$pval.cor)
#'     dcor <- -log10(dcmat$pvalue)
#'   } else if (criterion == "dcor") {
#'     if (is.null(dcmat$dcor) | is.null(dcmat$corr)) {
#'       stop("Distance correlation or correlation missing in dcmatrix object.")
#'     }
#'     corr <- abs(dcmat$corr)
#'     dcor <- dcmat$dcor
#'   }
#'   
#'   
#'   if (!dcmat$withY) {  
#'     lowtr <- lower.tri(corr)
#'     corr[lowtr] <- NA
#'     dcor[lowtr] <- NA
#'     Y <- X
#'   }
#'   
#'   selint <- which((dcor - corr) > diff, arr.ind=TRUE)
#'   
#'   return(selint)
#'  
#' }
#' 
#' #' @export
#' plotsubgr <- function(dcmat, candidates, data, cov.subgroup = names(data), splitnum = 3, method = "color", smooth = "lm", switchXY = FALSE, names = "plot", width = 480, height = 480) {
#' 
#'   if (is.null(dcmat$X)) {
#'     stop("Data needed to plot subgroup associations! Run dcmatrix again or add data manually to dcmat object.")
#'   } 
#'   
#'   if ("X" %in% cov.subgroup | "Y" %in% cov.subgroup)  {
#'     stop("Variable name X and Y not allowed in cov.subgroup. Please rename variables.")
#'   } 
#'   
#'   
#'   X <- as.matrix(dcmat$X)
#'   if (dcmat$withY) 
#'     Y <- as.matrix(dcmat$Y)
#'   
#'   if (length(candidates)==2)
#'     candidates <- matrix(candidates,ncol=2)
#'   
#'   names.Y <- "Y"
#'   
#'   if (is.null(colnames(X)))
#'     colnames(X) <- paste("X",1:ncol(X))
#'   
#' 
#' 
#'   
#'   
#'   if (!dcmat$withY) {  
#'     Y <- X
#'     names.Y <- "X"
#'     if (switchXY) {
#'       candidates <- cbind(candidates[,2],candidates[,1])
#'     }
#'   } else {
#'     if (dcmat$withY) {
#'       if (is.null(colnames(Y)))
#'         colnames(Y) <- paste("Y",1:ncol(Y), sep = "")
#'     }
#'    if (switchXY) {
#'       Xnew <- Y
#'       Ynew <- X
#'       X <- Xnew
#'       Y <- Ynew
#'       candidates <- cbind(candidates[,2],candidates[,1])
#'     }
#'   }
#'   
#'   nsubgr <- length(cov.subgroup)
#'   
#'   for (j in 1:nsubgr) {
#'     if (is.numeric(data[,cov.subgroup[j]])) {
#'       data[,cov.subgroup[j]] <- gtools::quantcut(data[,cov.subgroup[j]], q= splitnum)
#'     }
#'   }
#' 
#'   
#' 
#'   ncand <- nrow(candidates)
#'   for (i in 1:ncand) {
#'     for (j in 1:nsubgr) {
#'       name.X <- colnames(X)[candidates[i,1]]
#'       name.Y <- colnames(Y)[candidates[i,2]]
#'       
#'       title <- paste("Association of ", name.X," with ", name.Y, " in ",cov.subgroup[j]," subgroups",sep="")
#'       filename <- paste(names,"X-",name.X,"_Y-",name.Y,"_subgroup",cov.subgroup[j],".png",sep="")
#'       png(filename, width = width, height = height)
#'       plotobj <- ggplot2::ggplot(data)+ggtitle(title)+ ggplot2::xlab(name.X) + ggplot2::ylab(name.Y)
#'       
#'       plotobj <- plotobj + ggplot2::geom_point(ggplot2::aes(X[,candidates[i,1]], Y[,candidates[i,2]], color = data[,cov.subgroup[j]]))
#'       
#'       if (method == "facet")
#'         plotobj <- plotobj + ggplot2::geom_point(ggplot2::aes(X[,candidates[i,1]], Y[,candidates[i,2]])) + facet_grid(rows = vars(data[,cov.subgroup[j]]))
#'       
#'       if (smooth != "none")
#'         plotobj <- plotobj + geom_smooth(ggplot2::aes(X[,candidates[i,1]], Y[,candidates[i,2]], group=data[,cov.subgroup[j]], color = data[,cov.subgroup[j]]), method=smooth, se=F)  
#'       
#'       print(plotobj)
#'       
#'       dev.off()
#'     }
#'   }
#'   
#' }
#' 
#' is.dcmatrix <- function(x) inherits(x, "dcmatrix")

preptoterms.fast <- function(prepX, prepY, n, pairwise, discreteX, discreteY, perm) {
  terms <- list()
  
  if (!pairwise) {
    terms$aidot <- prepX$aidot
    terms$bidot <- prepY$aidot
    terms$adotdot <- prepX$adotdot
    terms$bdotdot <- prepY$adotdot
    terms$aijaij <- prepX$aijaij
    terms$bijbij <- prepY$aijaij
    terms$ncc <- n
  } else {
    cc <- intersect(prepX$cc, prepY$cc)
    ncc <- length(cc)
    terms$ncc <- ncc
  }
  
  if (!discreteX & !discreteY) {
    if (pairwise) {
  
      
      IX <- IY <- temp <- IY.s <-  1:ncc
      
      IX0here <- rep(NA,n)
      IXhere <- prepX$IX
      IXhere[prepY$mc] <- NA
      IX0here[IXhere[cc]] <- temp 
      cc0 <- !is.na(IX0here)
      IX0 <- IX0here[cc0]
      IX[IX0]<- temp
      
      IY0here <- rep(NA,n)
      IYhere <- prepY$IX
      IY0here[IYhere[cc]] <- temp 
      IYhere[prepX$mc] <- NA
      cc0 <- !is.na(IY0here)
      IY0 <- IY0here[cc0]
      IY[IY0]<- temp
      
      Xcc <- prepX$X[cc]
      Ycc <- prepY$X[cc]
      
      X.sort.X <- Xcc[IX0]
      Y.sort.Y <- Ycc[IY0]
      
      
      sX.X <- cumsum(X.sort.X)
      sY.Y <- cumsum(Y.sort.Y)
      
      alphaX <- IX - 1
      alphaY <- IY - 1
      betaX <- sX.X[IX] - X.sort.X[IX]
      betaY <- sY.Y[IY] - Y.sort.Y[IY]
      
      Xdot <- sX.X[ncc]
      Ydot <- sY.Y[ncc]
      
      terms$aidot <- Xdot + (2 * alphaX - ncc) * Xcc - 2 * betaX
      terms$bidot <- Ydot + (2 * alphaY - ncc) * Ycc - 2 * betaY
  
      terms$adotdot <- sum(terms$aidot)
      terms$bdotdot <- sum(terms$bidot)
      
      XXdot <- sum(Xcc^2)
      terms$aijaij <- 2 * ncc * XXdot - 2* Xdot^2
      
      YYdot <- sum(Ycc^2)
      terms$bijbij <- 2 * ncc * YYdot - 2* Ydot^2
      
  
      Y.sortX <- Ycc[IX0]
      XY.sortX <- Y.sortX * X.sort.X
      
      IY0.s <- IX[IY0]
      IY.s[IY0.s] <- temp
      
      
      
      
      } else {
      IY.s <- temp <- 1:n
   
      X.sort.X <- prepX$X.sort.X
      IX0 <- prepX$IX0
      
      Y.sortX <- prepY$X[IX0]
      XY.sortX <- Y.sortX *  X.sort.X
      
      IY0.s <- prepX$IX[prepY$IX0]
      IY.s[IY0.s] <- temp
 
      
      sX.X <- prepX$sX.X
      sY.Y <- prepY$sX.X
      
      Ycc <- prepY$X
      IY <- prepY$IX
   
      }
    
   sY.X <- cumsum(X.sort.X[IY0.s])
   sX.Y <- cumsum(Y.sortX)
   sY.XY <- cumsum(XY.sortX[IY0.s]) 
   sX.XY <- cumsum(XY.sortX)
      
   terms$aijbij <- SUMAIJBIJ(IY.s, X.sort.X, Y.sortX, XY.sortX, sX.X, sX.Y, sX.XY, sY.X[IY.s], sY.Y[IY.s], sY.XY[IY.s])
   
   if (perm) {
     terms$n <- terms$ncc
     terms$Y <- Ycc  
     terms$IX0 <- IX0
     terms$X.sort.X <- X.sort.X
     terms$IY <- IY
     terms$sX.X <- sX.X
     terms$sY.Y <- sY.Y
     terms$ndisc <- 0
   }

  
 } else if (discreteX + discreteY == 1) {
   if (discreteX) {
     prX <- prepY
     prY <- prepX
     prepX <- prX
     prepY <- prY
   }
   
   if (pairwise) {

     IX <- IY <- temp <- IY.s <-  1:ncc
     
     IX0here <- rep(NA,n)
     IXhere <- prepX$IX
     IXhere[prepY$mc] <- NA
     IX0here[IXhere[cc]] <- temp 
     cc0 <- !is.na(IX0here)
     IX0 <- IX0here[cc0]
     IX[IX0]<- temp
     
   
     Y <- prepY$X[cc]
     Y.sortX <- Y[IX0]
     X <- prepX$X[cc]
     X.sort.X <- X[IX0]
     
     sX.X <- cumsum(X.sort.X)
     alphaX <- 0:(ncc-1)
     betaX <- sX.X - X.sort.X
     Xdot <-sX.X[ncc]
     Y.sortX <- Y[IX0]
     
     terms$aidot <- Xdot + (2 * alphaX - ncc) * X.sort.X - 2 * betaX
     terms$adotdot <- sum(terms$aidot)
     XXdot <- sum(X^2)
     terms$aijaij <- 2 * ncc * XXdot - 2* Xdot^2

     nY <- table(Y)
     terms$bidot <- as.numeric(ncc - nY[Y.sortX])
     terms$bdotdot <- sum(terms$bidot)
     terms$bijbij <- ncc * (ncc - 1) - sum(nY * (nY - 1))
     
   } else {
     Y <- prepY$X
     IX0 <- prepX$IX0
     Y.sortX <- Y[IX0]
     X.sort.X <- prepX$X.sort.X
     nY <- table(Y)
     terms$aidot <- terms$aidot[IX0]
     terms$bidot <- terms$bidot[IX0]
   }
     
   levY <-levels(Y)
   aijbij <- terms$adotdot
     
   for (lvl in levY) {
     set0 <- which(Y.sortX == lvl)
     n0 <- length(set0)
     alphaX0 <- 0:(n0-1)
     X0 <- X.sort.X[set0]
     sX.X0 <- cumsum(X0)
     betaX0 <- sX.X0 - X0
     X0dot <- sX.X0[n0]
     aidot0 <- X0dot + (2 * alphaX0 - n0) * X0 - 2 * betaX0
     adotdot0 <- sum(aidot0)
     aijbij <- aijbij - adotdot0
   }
     
   terms$aijbij <- aijbij
   
   if (perm) {
     terms$X.sort.X <- X.sort.X
     terms$Y.sortX <- as.factor(Y.sortX)
     terms$levY <- levY
     terms$ndisc <- 1
   }
   
 } else if (discreteX & discreteY) {
      if (pairwise) {
        X <- prepX$X[cc]
        Y <- prepY$X[cc]
        nXY <- table(X,Y)
        nX <- rowSums(nXY)
        nY <- colSums(nXY)
        
        aidot <-ncc - nX
        terms$adotdot <- sum(aidot * nX)
        bidot <- ncc - nY
        terms$bdotdot <- sum(bidot * nY)
        
        terms$aidot <- aidot[X]
        terms$bidot <- bidot[Y] 
        terms$aijaij <- ncc * (ncc - 1) - sum(nX * (nX - 1)) 
        terms$bijbij <- ncc * (ncc - 1) - sum(nY * (nY - 1))
        
        terms$aijbij <- ncc * ncc - sum(nX * nX) - sum(nY * nY) + sum(nXY * nXY)
      
        if (perm) {
          terms$aidotshort <- aidot
          terms$bidotshort <- bidot
          terms$nX <- nX
          terms$nY <- nY
          terms$X <- X
          terms$Y <- Y
          terms$n <- terms$ncc
          terms$ndisc <- 2
        }   
      
    } else {
        nXY <- table(prepX$X,prepY$X)
        nX <- prepX$nX
        nY <- prepY$nX
        
        terms$aijbij <- n * n - sum(nX * nX) - sum(nY * nY) + sum(nXY * nXY)
        
        if (perm) {
          terms$aidotshort <- prepX$aidotshort
          terms$bidotshort <- prepY$aidotshort
          terms$nX <- prepX$nX
          terms$nY <- prepY$nX
          terms$X <- prepX$X
          terms$Y <- prepY$X
          terms$n <- terms$ncc
          terms$ndisc <- 2
        } 
        
    }
    
 }
 
  return(terms)
  
}




preptoterms.standard <- function(prepX, prepY, n, pairwise, perm) {
  distX <- prepX$distX
  distY <- prepY$distX

  if (!pairwise) {
    aijbij <- matrix_prod_sum(distX, distY)
    
    if (perm) {
      return(list("aijbij" = aijbij,"aijaij" = prepX$aijaij, "aidot" = prepX$aidot,"bijbij" = prepY$aijaij, "bidot" = prepY$aidot, "adotdot" = prepX$adotdot, "bdotdot" = prepY$adotdot, "ncc" = n, "distX" = distX, "distY" = distY))
    } else {
      return(list("aijbij" = aijbij,"aijaij" = prepX$aijaij, "aidot" = prepX$aidot,"bijbij" = prepY$aijaij, "bidot" = prepY$aidot, "adotdot" = prepX$adotdot, "bdotdot" = prepY$adotdot, "ncc" = n))
    }
  } else {
    cc <- intersect(prepX$cc, prepY$cc)
    ncc <- length(cc)
    aidot <- colsums_subset(distX,cc)
    adotdot <- sum(aidot)
    bidot <- colsums_subset(distY,cc)
    bdotdot <- sum(bidot)
    aijaij <- matrix_prod_sum_subset(distX, distX, cc)
    bijbij <- matrix_prod_sum_subset(distY, distY, cc)
    aijbij <- matrix_prod_sum_subset(distX, distY, cc)
    
    
    if (perm) {
      return(list("aijbij" = aijbij,"aijaij" = aijaij, "aidot" = aidot,"bijbij" = bijbij, "bidot" = bidot, "adotdot" = adotdot, "bdotdot" = bdotdot, "ncc" = ncc, "distX" = distX[cc,cc], "distY" = distY[cc,cc]))
    } else {
    return(list("aijbij" = aijbij, "aijaij" = aijaij, "bijbij" = bijbij, "aidot" = aidot, "bidot" = bidot, "adotdot" = adotdot, "bdotdot" = bdotdot, "ncc" = ncc))
    }
  }
}


preptoterms.memsave <- function(prepX, prepY, metr.X, metr.Y, n, pairwise, perm) {

  
   if (pairwise) {
    cc <- intersect(prepX$cc, prepY$cc)
    X <- prepX$X[cc]
    Y <- prepY$X[cc]
    ncc <- length(cc)
    dvartermsX <- dvarterms.memsave(X, metr.X, prepX$p)
    dvartermsY <- dvarterms.memsave(Y, metr.Y, prepY$p)
    aijbij <- aijbij.memsave(X, Y, metr.X, metr.Y, prepX$p, prepY$p)
 
    if (perm) { 
      dec.metr.X <- decom.metr(metr.X)
      metr.X <- dec.metr.X[1]
      prmX <- dec.metr.X[2]
      
      dec.metr.Y <- decom.metr(metr.Y)
      metr.Y <- dec.metr.Y[1]
      prmY <- dec.metr.Y[2]
      
      return(list("aijbij" = aijbij, "aijaij" = dvartermsX$aijaij, "bijbij" = dvartermsY$aijaij, "aidot" = dvartermsX$aidot, "bidot" = dvartermsY$aidot, "adotdot" = dvartermsX$adotdot, "bdotdot" = dvartermsY$adotdot, "ncc" = ncc, "X" = X, "Y" = Y, "p" = prepX$p, "q" = prepY$p, "metr.X" = metr.X, "metr.Y" = metr.X, "prmX" = prmX, "prmY" = prmY))
   }  else 
      return(list("aijbij" = aijbij, "aijaij" = dvartermsX$aijaij, "bijbij" = dvartermsY$aijaij, "aidot" = dvartermsX$aidot, "bidot" = dvartermsY$aidot, "adotdot" = dvartermsX$adotdot, "bdotdot" = dvartermsY$adotdot, "ncc" = ncc))
  } else {
    ncc <- n
    X <- prepX$X
    Y <- prepY$X

  aijbij <- aijbij.memsave(X, Y, metr.X, metr.Y, prepX$p, prepY$p)
  if (perm) {
    dec.metr.X <- decom.metr(metr.X)
    metr.X <- dec.metr.X[1]
    prmX <- dec.metr.X[2]
    
    dec.metr.Y <- decom.metr(metr.Y)
    metr.Y <- dec.metr.Y[1]
    prmY <- dec.metr.Y[2]
    return(list("aijbij" = aijbij, "aijaij" = prepX$aijaij, "aidot" = prepX$aidot, "bidot" = prepY$aidot, "bijbij" = prepY$aijaij, "adotdot" = prepX$adotdot, "bdotdot" = prepY$adotdot, "ncc" = ncc, "X" = X, "Y" = Y, "p" = prepX$p, "q" = prepY$p, "metr.X" = metr.X, "metr.Y" = metr.X, "prmX" = prmX, "prmY" = prmY))
  } else
    return(list("aijbij" = aijbij, "aijaij" = prepX$aijaij, "aidot" = prepX$aidot, "bidot" = prepY$aidot, "bijbij" = prepY$aijaij, "adotdot" = prepX$adotdot, "bdotdot" = prepY$adotdot, "ncc" = n, "X" = X, "Y" = Y, "p" = prepX$p, "q" = prepY$p, "metr.X" = metr.X, "metr.Y" = metr.X))
  }
}

dcovterms.smp <- function(terms, smp, alg.fast, alg.memsave, alg.standard, p, q, metr.X, metr.Y) {
  
  if (alg.fast) {
    if (p == 1 & q == 1) {
      if (metr.X == "euclidean" & metr.Y == "euclidean") {
          terms.smp <- function(terms, smp) {sampleterms.fast(terms, smp)}
      } else if (metr.X == "euclidean" & metr.Y == "discrete") {
        Y <- as.factor(Y)
        terms.smp <- function(terms, smp) {sampleterms.fast.numdisc(terms, smp)}
      } else if (metr.X == "discrete" & metr.Y == "euclidean") {
        X <- as.factor(X)
        terms.smp <- function(terms, smp) {sampleterms.fast.numdisc(terms, smp)}
      } else if (metr.X == "discrete" & metr.Y == "discrete") {
        X <- as.factor(X)
        Y <- as.factor(Y)
        terms.smp <- function(terms, smp) {sampleterms.fast.discrete(terms, smp)}
      } else {
        stop("metr.X and metr.Y have to be \"euclidean\" or \"discrete\" for fast algorithms")
      }
    } else {
      stop("Dimensions of X and Y must be 1 for fast algorithms.")
    }
  } else if (alg.memsave) {
    if (metr.X[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete") &
        metr.Y[1] %in% c("euclidean", "alpha", "gaussian", "boundsq", "minkowski", "discrete")) {
      terms.smp <- function(terms, smp) {sampleterms.memsave(terms, smp)}
    } else {
      stop("Memory efficient algorithms cannot be run with user-defined metrics")
    }
  } else if (alg.standard) {
    terms.smp <- function(terms, smp) {sampleterms.standard(terms, smp)}
  }
  return(terms.smp)  
}






sampleterms.standard <- function(terms, smp) {
  aijbij <- matrix_prod_sum_sample(terms$distX, terms$distY, smp)
  Sab <- vector_prod_sum_sample(terms$aidot, terms$bidot, smp)
  return(list("aijbij" = aijbij, "Sab" = Sab))
}


sampleterms.fast.matr <- function(terms, smp) {

 if (terms$ndisc == 0) 
   output <- sampleterms.fast(terms, smp)
 else if (terms$ndisc == 1)
   output <- sampleterms.fast.numdisc(terms, smp)
 else if (terms$ndisc == 2) {
    terms$aidot <- terms$aidotshort
    terms$bidot <- terms$bidotshort
    output <- sampleterms.fast.discrete(terms, smp)
 }
}


sampleterms.fast <- function(terms, smp) {
  IY0.s <- temp <-  1:terms$n
  Sab <- vector_prod_sum_sample(terms$aidot, terms$bidot, smp)
  
  Y.sortX <- terms$Y[smp][terms$IX0]
  XY.sortX <- Y.sortX * terms$X.sort.X
  
  IY.s <- terms$IY[smp][terms$IX0]
  IY0.s[IY.s] <- temp
  
  
  sY.X <- cumsum(terms$X.sort.X[IY0.s])
  sX.Y <- cumsum(Y.sortX)
  sY.XY <- cumsum(XY.sortX[IY0.s]) 
  sX.XY <- cumsum(XY.sortX)
  
  aijbij <- SUMAIJBIJ(IY.s, terms$X.sort.X, Y.sortX, XY.sortX, terms$sX.X, sX.Y, sX.XY, sY.X[IY.s], terms$sY.Y[IY.s], sY.XY[IY.s])
  
  return(list("aijbij" = aijbij, "Sab" = Sab))
}


sampleterms.fast.discrete <- function(terms, smp) {
  
  nXY <- table(terms$X,terms$Y[smp])
  aijbij <- terms$n * (terms$n - 1) - sum(terms$nX * (terms$nX - 1)) - sum(terms$nY * (terms$nY - 1)) + sum(nXY * (nXY - 1))
  Sab <- sum((terms$aidot %*% t(terms$bidot)) * nXY)
  

  return(list("aijbij" = aijbij, "Sab" = Sab))
  
}


sampleterms.fast.numdisc <- function(terms,smp) {
  aijbij <- terms$adotdot
  Y.sortX <- terms$Y.sortX[smp]
  
  
  for (lvl in terms$levY) {
    set0 <- which(Y.sortX == lvl)
    n0 <- length(set0)
    alphaX0 <- 0:(n0-1)
    X0 <- terms$X.sort.X[set0]
    sX.X0 <- cumsum(X0)
    betaX0 <- sX.X0 - X0
    X0dot <- sX.X0[n0]
    aidot0 <- X0dot + (2 * alphaX0 - n0) * X0 - 2 * betaX0
    adotdot0 <- sum(aidot0)
    aijbij <- aijbij - adotdot0
  }
  
  
  bidot <- terms$bidot[smp]

  Sab <- vector_prod_sum(terms$aidot,bidot)
  
  
  return(list("aijbij" = aijbij, "Sab" = Sab))
}


sampleterms.memsave <-  function(terms, smp) {
  

  if (terms$p == 1 & terms$q == 1)
    aijbij <- aijbijmemvec(terms$X, terms$Y[smp], terms$metr.X, terms$metr.Y, terms$prmX, terms$prmY)
    else if (terms$q ==1)
    aijbij <- aijbijmem(terms$X, as.matrix(terms$Y[smp,]), terms$metr.X, terms$metr.Y, terms$prmX, terms$prmY)
    else 
      aijbij <- aijbijmem(terms$X, terms$Y[smp,], terms$metr.X, terms$metr.Y, terms$prmX, terms$prmY)
    
    Sab <- vector_prod_sum_sample(terms$aidot,terms$bidot,smp)
    
    return(list("aijbij" = aijbij, "Sab" = Sab))
}  

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

squaresym <- function(A) {
    .Call(`_dcortools_squaresym`, A)
}

normalize_matrix <- function(M, cmM, mM) {
    .Call(`_dcortools_normalize_matrix`, M, cmM, mM)
}

hadamard_product <- function(X, Y) {
    .Call(`_dcortools_hadamard_product`, X, Y)
}

matrix_prod_sum2 <- function(X, Y) {
    .Call(`_dcortools_matrix_prod_sum2`, X, Y)
}

sum_hadamard_power2 <- function(X) {
    .Call(`_dcortools_sum_hadamard_power2`, X)
}

sum_hadamard_power3 <- function(X) {
    .Call(`_dcortools_sum_hadamard_power3`, X)
}

vector_product <- function(X, Y) {
    .Call(`_dcortools_vector_product`, X, Y)
}

matrix_prod_sum <- function(X, Y) {
    .Call(`_dcortools_matrix_prod_sum`, X, Y)
}

vector_prod_sum <- function(X, Y) {
    .Call(`_dcortools_vector_prod_sum`, X, Y)
}

specific_vector_prod_sum <- function(X, Y, gamma_1, gamma_X, gamma_Y, gamma_XY) {
    .Call(`_dcortools_specific_vector_prod_sum`, X, Y, gamma_1, gamma_X, gamma_Y, gamma_XY)
}

matrix_sum <- function(X) {
    .Call(`_dcortools_matrix_sum`, X)
}

matrix_prod_sum_sample <- function(X, Y, s) {
    .Call(`_dcortools_matrix_prod_sum_sample`, X, Y, s)
}

matrix_prod_sum_subset <- function(X, Y, s) {
    .Call(`_dcortools_matrix_prod_sum_subset`, X, Y, s)
}

colsums_subset <- function(X, s) {
    .Call(`_dcortools_colsums_subset`, X, s)
}

vector_prod_sum_sample <- function(X, Y, s) {
    .Call(`_dcortools_vector_prod_sum_sample`, X, Y, s)
}

dcovtermsmemvec <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0, calcdcor = FALSE, calcperm = FALSE) {
    .Call(`_dcortools_dcovtermsmemvec`, X, Y, metrX, metrY, prmX, prmY, calcdcor, calcperm)
}

dcovtermsmem <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0, calcdcor = FALSE, calcperm = FALSE) {
    .Call(`_dcortools_dcovtermsmem`, X, Y, metrX, metrY, prmX, prmY, calcdcor, calcperm)
}

dvartermsmemvec <- function(X, metrX, prmX = 0) {
    .Call(`_dcortools_dvartermsmemvec`, X, metrX, prmX)
}

dvartermsmem <- function(X, metrX, prmX = 0) {
    .Call(`_dcortools_dvartermsmem`, X, metrX, prmX)
}

aijbijmemvec <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0) {
    .Call(`_dcortools_aijbijmemvec`, X, Y, metrX, metrY, prmX, prmY)
}

aijbijmem <- function(X, Y, metrX, metrY, prmX = 0, prmY = 0) {
    .Call(`_dcortools_aijbijmem`, X, Y, metrX, metrY, prmX, prmY)
}

SUMAIJBIJ <- function(IY, X, Y, XY, SX_X, SX_Y, SX_XY, SY_X, SY_Y, SY_XY) {
    .Call(`_dcortools_SUMAIJBIJ`, IY, X, Y, XY, SX_X, SX_Y, SX_XY, SY_X, SY_Y, SY_XY)
}

#' @export
print.dctest <- function(x, ...) {
  dctest <- x
  if (dctest$method == "none")
    testout <- "None"
  else if (dctest$method == "permutation")
    testout <- paste("Permutation test with ",dctest$b," permutations")
  else if (dctest$method == "gamma")
    testout <- "Gamma moment matching"               
  else if (dctest$method == "conservativebb")
    testout <- "Two-moment approximation by Berschneider and Boettcher (2019). This method is typically conservative." 
  else if (dctest$method == "conservativesz")
    testout <- "Chi-Square approximation by Szekely, Rizzo, Bakirov (2007). This method is typically markedly conservative." 
  else if (dctest$method == "conservativespv")
    testout <- "Chi-Square approximation by Shen, Panda, Vogelstein (2021). This method is aymptotically conservative for p-values smaller than 0.215, but can be anti-conservative in small samples." 
  else if (dctest$method == "bb3")
    testout <- "Three moment approximation by Berschneider and Boettcher (2019)."
  else if (dctest$method == "spectral")
    testout <- "Spectral approach using the asymptotic distribution. Asymptotically valid and tight (Edelmann, 2026)."
  else if (dctest$method == "wildbs1")
    testout <- "Wild bootstrap by Chwialkowksi, et al., Method 1"
  else if (dctest$method == "wildbs2")
    testout <- "Wild bootstrap by Chwialkowksi, et al., Method 2"
  
  #if (dctest$pvalue<1e-6)
 #   pvalout <- "< 1E-6"
 # else
#pvalout <- round(dctest$pvalue,6)
  
  pvalout <- dctest$pvalue
  
  cat(paste("pvalue: ",pvalout,"   dcov: ",round(dctest$dcov,6),"   dcor: ",round(dctest$dcor,6)))
  cat("\n")
  cat("\n")
  cat(paste("Method:",testout, "\n") )
  cat("\n")
  cat("Call:", "\n")
  cat(paste(deparse(dctest$call), "\n"))
  cat("\n")
  cat(paste("Bias correction: ", ifelse(dctest$bias.corr,"Yes \n", "No \n")) )
  cat(paste("Affinely invariant: ", ifelse(dctest$affine,"Yes \n", "No \n")) )
  cat(paste("Standardization applied: ", ifelse(dctest$standardize,"Yes \n", "No \n")) )
  cat(paste("Metric for X: ", dctest$metr.X, "\n") )
  cat(paste("Metric for Y: ", dctest$metr.Y, "\n") )
}


is.dctest <- function(x) inherits(x, "dctest")

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distcov_function.R
\name{distcov}
\alias{distcov}
\title{Calculates the distance covariance \insertCite{szekely2007,szekely2009brownian}{dcortools}.}
\usage{
distcov(
  X,
  Y,
  affine = FALSE,
  standardize = FALSE,
  bias.corr = FALSE,
  type.X = "sample",
  type.Y = "sample",
  metr.X = "euclidean",
  metr.Y = "euclidean",
  use = "all",
  algorithm = "auto"
)
}
\arguments{
\item{X}{contains either the first  sample or its corresponding distance matrix.

In the first case, X can be provided either as a vector (if one-dimensional), a matrix or a data.frame (if two-dimensional or higher). 

In the second case, the input must be a distance matrix corresponding to the sample of interest.

If X is a sample, type.X must be specified as "sample". If X is a distance matrix, type.X must be specified as "distance".}

\item{Y}{see X.}

\item{affine}{logical; specifies if the affinely invariant distance covariance \insertCite{dueck2014affinely}{dcortools} should be calculated or not.}

\item{standardize}{logical; specifies if X and Y should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.}

\item{bias.corr}{logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.}

\item{type.X}{For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.}

\item{type.Y}{see type.X.}

\item{metr.X}{specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").

 Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
 
 For "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric, parameter)", e.g. c("gaussian", 3) for a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
 
 See \insertCite{lyons2013distance,sejdinovic2013equivalence,bottcher2017detecting;textual}{dcortools} for details.}

\item{metr.Y}{see metr.X.}

\item{use}{specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.}

\item{algorithm}{specifies the algorithm used for calculating the distance covariance. 

"fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 

"memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 

"standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.

"auto" chooses the best algorithm for the specific setting using a rule of thumb.}
}
\value{
numeric; the distance covariance between samples X and Y.
}
\description{
Calculates the distance covariance \insertCite{szekely2007,szekely2009brownian}{dcortools}.
}
\examples{
X <- rnorm(100)
Y <- X + 3 * rnorm(100)
distcov(X, Y) # standard distance covariance

distcov(X, Y, metr.X = "gaussauto", metr.Y = "gaussauto") 
# Gaussian distance with bandwidth choice based on median heuristic

distcov(X, Y, metr.X = c("alpha", 0.5), metr.Y = c("alpha", 0.5)) 
# alpha distance covariance with alpha = 0.5.


#Define a user-specified (slow) version of the alpha metric

alpha_user <- function(X, prm = 1, kernel = FALSE) {
    as.matrix(dist(X)) ^ prm
}

distcov(X, Y, metr.X = c("alpha", 0.5), metr.Y = c("alpha", 0.5)) 
# Gives the same result as before.
   

#User-specified Gaussian kernel function  
     
gauss_kernel <- function(X, prm = 1, kernel = TRUE)  {
    exp(as.matrix(dist(X)) ^ 2 / 2 / prm ^ 2)
}  

distcov(X, Y, metr.X = c("gauss_kernel", 2), metr.Y = c("gauss_kernel", 2)) 
# calculates the distance covariance using the corresponding kernel-induced metric

distcov(X, Y, metr.X = c("gaussian", 2), metr.Y = c("gaussian", 2)) 
# same result

Y <- matrix(nrow = 100, ncol = 2)
X <- rnorm(300)
dim(X) <- c(100, 3)
Z <- rnorm(100)
Y <- matrix(nrow = 100, ncol = 2)
Y[, 1] <- X[, 1] + Z
Y[, 2] <- 3 * Z

distcov(X, Y) 

distcov(X, Y, affine = TRUE) 
# affinely invariant distance covariance

distcov(X, Y, standardize = TRUE) 
## distance covariance standardizing the components of X and Y

}
\references{
\insertRef{bottcher2017detecting}{dcortools}

\insertRef{dueck2014affinely}{dcortools}

\insertRef{huo2016fast}{dcortools}

\insertRef{lyons2013distance}{dcortools}

\insertRef{sejdinovic2013equivalence}{dcortools}

\insertRef{szekely2007}{dcortools}

\insertRef{szekely2009brownian}{dcortools}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auxiliary_functions.R
\name{extract_np}
\alias{extract_np}
\title{Extract the dimensions of X.}
\usage{
extract_np(X, type.X)
}
\arguments{
\item{X}{a numeric vector or a numeric matrix.}

\item{type.X}{either "sample" or "distance". If type.X = "sample", X must be
a numeric vector or numeric matrix with the corresponding observations. If metr.X = "distance",
X must be a distance matrix.}
}
\value{
The centralized distance matrix corresponding to X.
}
\description{
Extract the dimensions of X.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_dcmatrix.R
\name{hsplot}
\alias{hsplot}
\title{Plots Pearson/Spearman/Kendall correlation against distance correlation (often resembling a horseshoe(hs)).}
\usage{
hsplot(dcmat, maxcomp = 1e+05, col = "blue", alpha = 1, cortrafo = "none")
}
\arguments{
\item{dcmat}{A dcmatrix object.}

\item{maxcomp}{Maximum number of associations, for which distance correlation is plotted against correlation. If the number of associations in the dcmat object is larger, only the maxcomp associations with the largest difference between distance correlation and absolute (Pearson/Spearman/Kendall) correlation are plotted.}

\item{col}{color of the plot.}

\item{alpha}{alpha parameter of the plot.}

\item{cortrafo}{Either "none" or "gaussiandcor". If "gaussiandcor", the distance correlation under assumption of normality is calculated and plotted against the actual distance correlation. 

Note that this is only sensible for Pearson correlation!}
}
\value{
Plot of (possibly transformed) Pearson/Spearman/Kendall correlation against distance correlation.
}
\description{
Plots Pearson/Spearman/Kendall correlation against distance correlation (often resembling a horseshoe(hs)).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distcovsurv.R
\name{ipcw.dcov.test}
\alias{ipcw.dcov.test}
\title{Performs a permutation test based on the IPCW distance covariance.}
\usage{
ipcw.dcov.test(
  Y,
  X,
  affine = FALSE,
  standardize = FALSE,
  timetrafo = "none",
  type.X = "sample",
  metr.X = "euclidean",
  use = "all",
  cutoff = NULL,
  B = 499
)
}
\arguments{
\item{Y}{A column with two rows, where the first row contains the survival times and the second row the status indicators (a survival object will work).}

\item{X}{A vector or matrix containing the covariate information.}

\item{affine}{logical; indicates if X should be transformed such that the result is invariant under affine transformations of X.}

\item{standardize}{logical; should X be standardized using the standard deviations of single observations. No effect when affine = TRUE.}

\item{timetrafo}{specifies a transformation applied on the follow-up times. Can be "none", "log" or a user-specified function.}

\item{type.X}{For "distance", X is interpreted as a distance matrix. For "sample" (or any other value), X is interpreted as a sample.}

\item{metr.X}{metr.X specifies the metric which should be used for X to analyze the distance covariance. Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto" and "boundsq". For "alpha", "minkowski", "gauss", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric,parameter)" (see examples); the standard parameter is 2 for "minkowski" and "1" for all other metrics.}

\item{use}{specifies how to treat missing values. "complete.obs" excludes observations containing NAs, "all" uses all observations.}

\item{cutoff}{If provided, all survival times larger than cutoff are set to the cutoff and all corresponding status indicators are set to one. Under most circumstances, choosing a cutoff is highly recommended.}

\item{B}{The number of permutations used for the permutation test}
}
\value{
An list with two arguments, $dcov contains the IPCW distance covariance, $pvalue the corresponding p-value
}
\description{
Performs a permutation test based on the IPCW distance covariance.
}
\examples{
X <- rnorm(100)
survtime <- rgamma(100, abs(X))
cens <- rexp(100)
status <- as.numeric(survtime < cens)
time <- sapply(1:100, function(u) min(survtime[u], cens[u]))
surv <- cbind(time, status)
ipcw.dcov.test(surv, X)
ipcw.dcov.test(surv, X, cutoff = quantile(time, 0.8)) 
# often better performance when using a cutoff time

}
\references{
\insertRef{bottcher2017detecting}{dcortools}

\insertRef{datta2010inverse}{dcortools}

\insertRef{dueck2014affinely}{dcortools}

\insertRef{huo2016fast}{dcortools}

\insertRef{lyons2013distance}{dcortools}

\insertRef{sejdinovic2013equivalence}{dcortools}

\insertRef{szekely2007}{dcortools}

\insertRef{szekely2009brownian}{dcortools}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dc_matrix2.R
\name{dcmatrix}
\alias{dcmatrix}
\title{Calculates distance covariance and distance correlation matrices}
\usage{
dcmatrix(
  X,
  Y = NULL,
  calc.dcov = TRUE,
  calc.dcor = TRUE,
  calc.cor = "none",
  calc.pvalue.cor = FALSE,
  return.data = TRUE,
  test = "none",
  adjustp = "none",
  b = 499,
  affine = FALSE,
  standardize = FALSE,
  bias.corr = FALSE,
  group.X = NULL,
  group.Y = NULL,
  metr.X = "euclidean",
  metr.Y = "euclidean",
  use = "all",
  algorithm = "auto",
  fc.discrete = FALSE,
  calc.dcor.pw = FALSE,
  calc.dcov.pw = FALSE,
  test.pw = "none",
  metr.pw.X = "euclidean",
  metr.pw.Y = "euclidean"
)
}
\arguments{
\item{X}{A data.frame or matrix.}

\item{Y}{Either NULL or a data.frame or a matrix with the same number of rows as X. If only X is provided, distance covariances/correlations are calculated between all groups in X. If X and Y are provided, distance covariances/correlations are calculated between all groups in X and all groups of Y.}

\item{calc.dcov}{logical; specifies if the distance covariance matrix is calculated.}

\item{calc.dcor}{logical; specifies if the distance correlation matrix is calculated.}

\item{calc.cor}{If set as "pearson", "spearman" or "kendall", a corresponding correlation matrix is additionally calculated.}

\item{calc.pvalue.cor}{logical; IF TRUE, a p-value based on the Pearson or Spearman correlation matrix is calculated (not implemented for calc.cor ="kendall") using Hmisc::rcorr.}

\item{return.data}{logical; specifies if the dcmatrix object should contain the original data.}

\item{test}{specifies the type of test that is performed, "permutation" performs a Monte Carlo Permutation test. "gamma" performs a test based on a gamma approximation of the test statistic under the null. "conservative" performs a conservative two-moment approximation. "bb3" performs a quite precise three-moment approximation and is recommended when computation time is not an issue.}

\item{adjustp}{If setting this parameter to "holm", "hochberg", "hommel", "bonferroni", "BH", "BY" or "fdr", corresponding adjusted p-values are additionally returned for the distance covariance test.}

\item{b}{specifies the number of random permutations used for the permutation test. Ignored for all other tests.}

\item{affine}{logical; indicates if the affinely transformed distance covariance should be calculated or not.}

\item{standardize}{specifies if data should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.}

\item{bias.corr}{logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.}

\item{group.X}{A vector, each entry specifying the group membership of the respective column in X. Each group is handled as one sample for calculating the distance covariance/correlation matrices. If NULL, every sample is handled as an individual group.}

\item{group.Y}{A vector, each entry specifying the group membership of the respective column in Y. Each group is handled as one sample for calculating the distance covariance/correlation matrices. If NULL, every sample is handled as an individual group.}

\item{metr.X}{Either a single metric or a list providing a metric for each group in X (see examples).}

\item{metr.Y}{see metr.X.}

\item{use}{"all" uses all observations, "complete.obs" excludes NAs, "pairwise.complete.obs" uses pairwise complete observations for each comparison.}

\item{algorithm}{specifies the algorithm used for calculating the distance covariance. 

"fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 

"memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 

"standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.

"auto" chooses the best algorithm for the specific setting using a rule of thumb.

"memsave" is typically very inefficient for dcmatrix and should only be applied in exceptional cases.}

\item{fc.discrete}{logical; If TRUE, "discrete" metric is applied automatically on samples of type "factor" or "character".}

\item{calc.dcor.pw}{logical; If TRUE, a distance correlation matrix between the univariate observations/columns is additionally calculated. Not meaningful if group.X and group.Y are not specified.}

\item{calc.dcov.pw}{logical; If TRUE, a distance covariance matrix between the univariate observations/columns is additionally calculated. Not meaningful if group.X and group.Y are not specified.}

\item{test.pw}{specifies a test (see argument "test") that is performed between all single observations.}

\item{metr.pw.X}{Either a single metric or a list providing a metric for each single observation/column in X (see metr.X).}

\item{metr.pw.Y}{See metr.pw.Y.}
}
\value{
S3 object of class "dcmatrix" with the following components
\item{name X, Y}{description original data (if return.data = TRUE).} 
\item{name dcov, dcor}{distance covariance/correlation matrices between the groups specified in group.X/group.Y (if calc.dcov/calc.dcor = TRUE).} 
\item{name corr}{correlation matrix between the univariate observations/columns (if cal.cor is "pearson", "spearman" or "kendall").}
\item{name pvalue}{matrix of p-values based on a corresponding distance covariance test based on the entries in dcov (if argument test is not "none").} 
\item{name pvalue.adj}{matrix of p-values adjusted for multiple comparisons using the method specified in argument adjustp.} 
\item{name pvalue.cor}{matrix of pvalues based on "pearson"/"spearman" correlation (if calc.cor is "pearson" or "spearman" and calc.pvalue.cor = TRUE).}
\item{name dcov.pw,dcor.pw}{distance covariance/correlation matrices between the univariate observations (if calc.dcov.pw/calc.dcor.pw = TRUE.)} 
\item{name pvalue.pw}{matrix of p-values based on a corresponding distance covariance test based on the entries in dcov.pw (if argument test is not "none").}
}
\description{
Calculates distance covariance and distance correlation matrices
}
\examples{
X <- matrix(rnorm(1000), ncol = 10)

dcm <- dcmatrix(X, test="bb3",calc.cor = "pearson",
 calc.pvalue.cor = TRUE, adjustp = "BH") 
 
dcm <- dcmatrix(X, test="bb3",calc.cor = "pearson", 
 calc.pvalue.cor = TRUE, adjustp = "BH", 
 group.X = c(rep(1, 5), rep(2, 5)), 
 calc.dcor.pw = TRUE, test.pw = "bb3")


Y <- matrix(rnorm(600), ncol = 6)

Y[,6] <- rbinom(100, 4, 0.3)

dcm <- dcmatrix(X, Y, test="bb3",calc.cor = "pearson",
 calc.pvalue.cor = TRUE, adjustp = "BH")
  
dcm <- dcmatrix(X, Y, test="bb3",calc.cor = "pearson",
 calc.pvalue.cor = TRUE, adjustp = "BH",
 group.X = c(rep("group1", 5), rep("group2", 5)),
 group.Y = c(rep("group1", 5), "group2"), 
 metr.X = "gaussauto",
 metr.Y = list("group1" = "gaussauto", "group2" = "discrete"))
}
\references{
\insertRef{berschneider2018complex}{dcortools}

\insertRef{bottcher2017detecting}{dcortools}

\insertRef{dueck2014affinely}{dcortools}

\insertRef{huang2017statistically}{dcortools}

\insertRef{huo2016fast}{dcortools}

\insertRef{lyons2013distance}{dcortools}

\insertRef{sejdinovic2013equivalence}{dcortools}

\insertRef{szekely2007}{dcortools}

\insertRef{szekely2009brownian}{dcortools}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_dcmatrix.R
\name{dcorgaussianbiv}
\alias{dcorgaussianbiv}
\title{Calculates distance correlation from Pearson correlation under assumption of a bivariate normal distribution}
\usage{
dcorgaussianbiv(rho)
}
\arguments{
\item{rho}{Pearson correlation.}
}
\value{
Distance correlation assuming a bivariate normal distribution
}
\description{
Calculates distance correlation from Pearson correlation under assumption of a bivariate normal distribution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distcov_test2.R
\name{distcov.test}
\alias{distcov.test}
\title{Performs a distance covariance test.}
\usage{
distcov.test(
  X,
  Y,
  method = "permutation",
  b = 499L,
  ln = 20,
  affine = FALSE,
  standardize = FALSE,
  bias.corr = FALSE,
  type.X = "sample",
  type.Y = "sample",
  metr.X = "euclidean",
  metr.Y = "euclidean",
  use = "all",
  return.data = FALSE,
  algorithm = "auto",
  settings = NULL
)
}
\arguments{
\item{X}{contains either the first sample or its corresponding distance matrix.

In the first case, X can be provided either as a vector (if one-dimensional), a matrix or a data.frame (if two-dimensional or higher). 

In the second case, the input must be a distance matrix corresponding to the sample of interest.

If X is a sample, type.X must be specified as "sample". If X is a distance matrix, type.X must be specified as "distance".}

\item{Y}{see X.}

\item{method}{specifies the type of test that is performed.
 
"permutation" performs a Monte Carlo Permutation test. 

"gamma" performs a test based on a gamma approximation of the test statistic under the null \insertCite{huang2017statistically}{dcortools}. This test tends to be anti-conservative, if the ``real'' p-value is small

"conservativebb" performs an (asymptotically) conservative two-moment approximation \insertCite{berschneider2018complex}{dcortools}.

"conservativesz" performs the (asymptotically) conservative procedure proposed by \insertCite{szekely2007}{dcortools}.

"spv" performs the Chi-Square Test of Distance correlation proposed by Shen, Panda and Vogelstein (2021).

"spectral" performs the spectral approach based on the asymptotic distribution.
  
"bb3" performs a  three-moment approximation  \insertCite{berschneider2018complex}{dcortools}. This is the most precise parametric option, but only available with the standard algorithm.

"wildbs1" and "wildbs2" perform wild bootstrap tests \insertCite{chwialkowski2014wild}{dcortools}; experimental at the moment.}

\item{b}{integer; specifies the number of random permutations/bootstrap samples used for the permutation or wild bootstraps tests. Ignored for other tests.}

\item{ln}{numeric; block size parameter for wild bootstrap tests. Ignored for other tests.}

\item{affine}{logical; specifies if the affinely invariant distance covariance \insertCite{dueck2014affinely}{dcortools} should be calculated or not.}

\item{standardize}{logical; specifies if X and Y should be standardized dividing each component by its standard deviations. No effect when affine = TRUE.}

\item{bias.corr}{logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.}

\item{type.X}{For "distance", X is interpreted as a distance matrix. For "sample", X is interpreted as a sample.}

\item{type.Y}{see type.X.}

\item{metr.X}{specifies the metric which should be used to compute the distance matrix for X (ignored when type.X = "distance").

 Options are "euclidean", "discrete", "alpha", "minkowski", "gaussian", "gaussauto", "boundsq" or user-specified metrics (see examples).
 
 For "alpha", "minkowski", "gauss", "gaussauto" and "boundsq", the corresponding parameters are specified via "c(metric, parameter)", c("gaussian", 3) for example uses a Gaussian metric with bandwidth parameter 3; the default parameter is 2 for "minkowski" and "1" for all other metrics.
 
 See \insertCite{lyons2013distance,sejdinovic2013equivalence,bottcher2017detecting;textual}{dcortools} for details.}

\item{metr.Y}{see metr.X.}

\item{use}{specifies how to treat missing values. "complete.obs" excludes NAs, "all" uses all observations.}

\item{return.data}{logical; specifies if the test object should contain the original data.}

\item{algorithm}{specifies the algorithm used for calculating the distance covariance.}

\item{settings}{a named list of additional tuning parameters, used to override defaults for specific tests. For the spectral test ("spectral") this includes \code{starteig}, \code{multeig}, \code{large}, \code{tolfac}, \code{nomlevel}, \code{maxlz} and \code{weedlimit}; \code{exactmom} controls exact-moment computation. Mainly for advanced or internal use; if \code{NULL}, sensible defaults are chosen.
"fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 

"memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 

"standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.

"auto" chooses the best algorithm for the specific setting using a rule of thumb.}
}
\value{
distcov.test object
}
\description{
Performs a distance covariance test.
}
\references{
\insertRef{berschneider2018complex}{dcortools}

\insertRef{bottcher2017detecting}{dcortools}

\insertRef{chwialkowski2014wild}{dcortools}

\insertRef{dueck2014affinely}{dcortools}

\insertRef{huang2017statistically}{dcortools}

\insertRef{huo2016fast}{dcortools}

\insertRef{lyons2013distance}{dcortools}

\insertRef{sejdinovic2013equivalence}{dcortools}

\insertRef{szekely2007}{dcortools}

\insertRef{szekely2009brownian}{dcortools}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_dcmatrix.R
\name{plot.dcmatrix}
\alias{plot.dcmatrix}
\title{Plots a heatmap from a dcmatrix object using the function "pheatmap" from the package "pheatmap".}
\usage{
\method{plot}{dcmatrix}(
  x,
  type = "dcor",
  trunc.up = NULL,
  trunc.low = NULL,
  cluster_rows = FALSE,
  cluster_cols = FALSE,
  display_numbers = TRUE,
  ...
)
}
\arguments{
\item{x}{a dcmatrix object.}

\item{type}{specifies what should be displayed in the heatmap. One of "dcor", "dcov", "logp" (-log10 of corresponding p-values), "cor", "abscor" (absolute correlation), "logp.cor", "dcor.pw", "dcov.pw" or "logp.pw".}

\item{trunc.up}{truncates the values to be plotted; if set to numeric, all values larger than trunc.up are set to trunc.up.}

\item{trunc.low}{truncates the values to be plotted; if set to numeric, all values smaller than trunc.low are set to trunc.low.}

\item{cluster_rows, cluster_cols, display_numbers}{passed to pheatmap().}

\item{...}{passed to pheatmap().}
}
\value{
a heatmap plotting the entries of the slot specified in type of the object specified in dcmat.
}
\description{
Plots a heatmap from a dcmatrix object using the function "pheatmap" from the package "pheatmap".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dcsis.R
\name{dcsis}
\alias{dcsis}
\title{Performs distance correlation sure independence screening \insertCite{li2012feature}{dcortools} with some additional options (such as calculating corresponding tests).}
\usage{
dcsis(
  X,
  Y,
  k = floor(nrow(X)/log(nrow(X))),
  threshold = NULL,
  calc.cor = "spearman",
  calc.pvalue.cor = FALSE,
  return.data = FALSE,
  test = "none",
  adjustp = "none",
  b = 499,
  bias.corr = FALSE,
  use = "all",
  algorithm = "auto"
)
}
\arguments{
\item{X}{A dataframe or matrix.}

\item{Y}{A vector-valued response having the same length as the number of rows of X.}

\item{k}{Number of variables that are selected (only used when threshold is not provided).}

\item{threshold}{If provided, variables with a distance correlation larger than threshold are selected.}

\item{calc.cor}{If set as "pearson", "spearman" or "kendall", a corresponding correlation matrix is additionally calculated.}

\item{calc.pvalue.cor}{logical; IF TRUE, a p-value based on the Pearson or Spearman correlation matrix is calculated (not implemented for calc.cor = "kendall") using Hmisc::rcorr.}

\item{return.data}{logical; specifies if the dcmatrix object should contain the original data.}

\item{test}{Allows for additionally calculating a test based on distance Covariance. Specifies the type of test that is performed, "permutation" performs a Monte Carlo Permutation test. "gamma" performs a test based on a gamma approximation of the test statistic under the null. "conservative" performs a conservative two-moment approximation. "bb3" performs a quite precise three-moment approximation and is recommended when computation time is not an issue.}

\item{adjustp}{If setting this parameter to "holm", "hochberg", "hommel", "bonferroni", "BH", "BY" or "fdr", corresponding adjusted p-values are additionally returned for the distance covariance test.}

\item{b}{specifies the number of random permutations used for the permutation test. Ignored for all other tests.}

\item{bias.corr}{logical; specifies if the bias corrected version of the sample distance covariance \insertCite{huo2016fast}{dcortools} should be calculated.}

\item{use}{"all" uses all observations, "complete.obs" excludes NAs, "pairwise.complete.obs" uses pairwise complete observations for each comparison.}

\item{algorithm}{specifies the algorithm used for calculating the distance covariance. 

"fast" uses an O(n log n) algorithm if the observations are one-dimensional and metr.X and metr.Y are either "euclidean" or "discrete", see also \insertCite{huo2016fast;textual}{dcortools}. 

"memsave" uses a memory saving version of the standard algorithm with computational complexity O(n^2) but requiring only O(n) memory. 

"standard" uses the classical algorithm. User-specified metrics always use the classical algorithm.

"auto" chooses the best algorithm for the specific setting using a rule of thumb.

"memsave" is typically very inefficient for dcsis and should only be applied in exceptional cases.}
}
\value{
dcmatrix object with the following two additional slots:
\item{name selected}{description indices of selected variables.} 
\item{name dcor.selected}{distance correlation of the selected variables and the response Y.}
}
\description{
Performs distance correlation sure independence screening \insertCite{li2012feature}{dcortools} with some additional options (such as calculating corresponding tests).
}
\examples{
X <- matrix(rnorm(1e5), ncol = 1000)
Y <- sapply(1:100, function(u) sum(X[u, 1:50])) + rnorm(100)
a <- dcsis(X, Y)
}
\references{
\insertRef{berschneider2018complex}{dcortools}
\insertRef{dueck2014affinely}{dcortools}

\insertRef{huang2017statistically}{dcortools}

\insertRef{huo2016fast}{dcortools}

\insertRef{li2012feature}{dcortools}

\insertRef{szekely2007}{dcortools}

\insertRef{szekely2009brownian}{dcortools}
}
